import json
from textwrap import dedent
from typing import Optional
from unittest import TestCase, mock

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.node.command import (
    node_attribute_output_cmd,
    node_utilization_output_cmd,
)
from pcs.common.interface import dto
from pcs.common.pacemaker.cluster_property import ClusterPropertyMetadataDto
from pcs.common.pacemaker.node import CibNodeListDto
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
    ListCibNvsetDto,
)

from pcs_test.tools.custom_mock import RuleInEffectEvalMock
from pcs_test.tools.misc import dict_to_modifiers
from pcs_test.tools.nodes_dto import get_nodes_dto

FIXTURE_NODE_CONFIG = get_nodes_dto(RuleInEffectEvalMock({}))

UTILIZATION_WARNING = (
    "Utilization attributes configuration has no effect until cluster property "
    "option 'placement-strategy' is set to one of the values: 'balanced', "
    "'minimal', 'utilization'"
)

FIXTURE_PLACEMENT_STRATEGY_SET = ListCibNvsetDto(
    nvsets=[
        CibNvsetDto(
            id="cib-bootstrap-options",
            options={},
            rule=None,
            nvpairs=[
                CibNvpairDto(
                    id="id-ps-b", name="placement-strategy", value="balanced"
                )
            ],
        )
    ]
)


FIXTURE_TEXT_OUTPUT = {
    "utilization": {
        "cmd": (
            "pcs -- node utilization node1 cpu=4 ram=32;\n"
            "pcs -- node utilization node2 cpu=8 ram=64"
        ),
        "text": dedent(
            """\
                Node: node1
                  Description: node1 desc
                  Utilization: nodes-1-utilization score=50
                    cpu=4
                    ram=32
                Node: node2
                  Utilization: nodes-2-utilization
                    cpu=8
                    ram=64
                    Rule: boolean-op=and score=INFINITY
                      Expression: date gt 2000-01-01"""
        ),
        "node": dedent(
            """\
                Node: node2
                  Utilization: nodes-2-utilization
                    cpu=8
                    ram=64
                    Rule: boolean-op=and score=INFINITY
                      Expression: date gt 2000-01-01"""
        ),
        "name": dedent(
            """\
                Node: node1
                  Description: node1 desc
                  Utilization: nodes-1-utilization score=50
                    cpu=4
                Node: node2
                  Utilization: nodes-2-utilization
                    cpu=8
                    Rule: boolean-op=and score=INFINITY
                      Expression: date gt 2000-01-01"""
        ),
        "node_and_name": dedent(
            """\
                Node: node2
                  Utilization: nodes-2-utilization
                    ram=64
                    Rule: boolean-op=and score=INFINITY
                      Expression: date gt 2000-01-01"""
        ),
    },
    "attribute": {
        "cmd": (
            "pcs -- node attribute node1 a=1 b=2;\n"
            "pcs -- node attribute node2 a=1 b=2"
        ),
        "text": dedent(
            """\
                Node: node1
                  Description: node1 desc
                  Attributes: nodes-1 score=50
                    a=1
                    b=2
                    Rule: score=INFINITY
                      Expression: date lt 2000-01-01
                Node: node2
                  Attributes: nodes-2
                    a=1
                    b=2"""
        ),
        "node": dedent(
            """\
                Node: node2
                  Attributes: nodes-2
                    a=1
                    b=2"""
        ),
        "name": dedent(
            """\
                Node: node1
                  Description: node1 desc
                  Attributes: nodes-1 score=50
                    a=1
                    Rule: score=INFINITY
                      Expression: date lt 2000-01-01
                Node: node2
                  Attributes: nodes-2
                    a=1"""
        ),
        "node_and_name": dedent(
            """\
                Node: node2
                  Attributes: nodes-2
                    b=2"""
        ),
    },
}


class NodeOutputCmdBaseMixin:
    def _assert_lib_mocks(self, utilization_warning=False):
        self.node.get_config_dto.assert_called_once_with()
        if self.command == "utilization":
            if utilization_warning:
                self.cluster_property.get_properties.assert_called_once_with()
                self.cluster_property.get_properties_metadata.assert_called_once_with()
            else:
                self.cluster_property.get_properties.assert_not_called()
                self.cluster_property.get_properties_metadata.assert_not_called()

    def test_no_text_output_format_and_node_arg(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["node"], {"output-format": "json"})
        self.assertEqual(
            cm.exception.message,
            "filtering is not supported with --output-format=cmd|json",
        )
        mock_print.assert_not_called()

    def test_no_text_output_format_and_name_option(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"name": "attr-name", "output-format": "cmd"})
        self.assertEqual(
            cm.exception.message,
            "filtering is not supported with --output-format=cmd|json",
        )
        mock_print.assert_not_called()

    def test_unknown_option(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"all": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--all' is not supported in this command",
        )
        mock_print.assert_not_called()

    def test_more_than_one_arg(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["node1", "node2"])
        self.assertIsNone(cm.exception.message)
        mock_print.assert_not_called()

    def test_unsupported_output_format(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"output-format": "unsupported"})
        self.assertEqual(
            cm.exception.message,
            (
                "Unknown value 'unsupported' for '--output-format' option. "
                "Supported values are: 'cmd', 'json', 'text'"
            ),
        )
        self.node.get_config_dto.assert_not_called()
        mock_print.assert_not_called()

    def test_empty_config_default_output_format(self, mock_print):
        self._set_empty_config()
        self._call_cmd([])
        self._assert_lib_mocks(utilization_warning=True)
        mock_print.assert_not_called()

    def test_empty_config_json_output_format(self, mock_print):
        self._set_empty_config()
        self._call_cmd([], {"output-format": "json"})
        self._assert_lib_mocks()
        mock_print.assert_called_once_with('{"nodes": []}')

    def test_empty_config_cmd_output_format(self, mock_print):
        self._set_empty_config()
        self._call_cmd([], {"output-format": "cmd"})
        self._assert_lib_mocks()
        mock_print.assert_not_called()

    def test_output_format_json(self, mock_print):
        self._call_cmd([], {"output-format": "json"})
        self._assert_lib_mocks()
        mock_print.assert_called_once_with(
            json.dumps(dto.to_dict(FIXTURE_NODE_CONFIG))
        )

    def test_output_format_cmd(self, mock_print):
        self._call_cmd([], {"output-format": "cmd"})
        self._assert_lib_mocks()
        mock_print.assert_called_once_with(
            FIXTURE_TEXT_OUTPUT[self.command]["cmd"]
        )

    def test_output_format_text(self, mock_print):
        self._call_cmd([], {"output-format": "text"})
        self._assert_lib_mocks(utilization_warning=True)
        mock_print.assert_called_once_with(
            FIXTURE_TEXT_OUTPUT[self.command]["text"]
        )

    def test_output_format_text_filter_node(self, mock_print):
        self._call_cmd(["node2"])
        self._assert_lib_mocks(utilization_warning=True)
        mock_print.assert_called_once_with(
            FIXTURE_TEXT_OUTPUT[self.command]["node"]
        )

    def test_output_format_text_filter_node_does_not_exist(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["nodeX"])
        self.assertEqual(cm.exception.message, "Unable to find a node: nodeX")
        self._assert_lib_mocks()
        mock_print.assert_not_called()

    def test_output_format_text_filter_name(self, mock_print):
        self._call_cmd([], {"name": self.filter_name})
        self._assert_lib_mocks(utilization_warning=True)
        mock_print.assert_called_once_with(
            FIXTURE_TEXT_OUTPUT[self.command]["name"]
        )

    def test_output_format_text_filter_node_and_name(self, mock_print):
        self._call_cmd(["node2"], {"name": self.filter_node_and_name})
        self._assert_lib_mocks(utilization_warning=True)
        mock_print.assert_called_once_with(
            FIXTURE_TEXT_OUTPUT[self.command]["node_and_name"]
        )


@mock.patch("pcs.cli.node.command.print")
class TestNodeAttributeOutputCmd(NodeOutputCmdBaseMixin, TestCase):
    command = "attribute"
    filter_name = "a"
    filter_node_and_name = "b"
    label = "Attributes"
    set_id = "ia"

    def setUp(self):
        self.lib = mock.Mock(spec_set=["node"])
        self.node = mock.Mock(spec_set=["get_config_dto"])
        self.lib.node = self.node
        self._set_config()

    def _call_cmd(self, argv, modifiers=None):
        node_attribute_output_cmd(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def _set_config(
        self,
        node_config: Optional[CibNodeListDto] = None,
    ):
        self.node.get_config_dto.return_value = FIXTURE_NODE_CONFIG
        if node_config:
            self.node.get_config_dto.return_value = node_config

    def _set_empty_config(self):
        self.node.get_config_dto.return_value = CibNodeListDto(nodes=[])

    def test_force_option_shows_help(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"force": True})
        self.assertIsNone(cm.exception.message)
        mock_print.assert_not_called()


@mock.patch("pcs.cli.node.command.print")
class TestNodeUtilizationOutputCmd(NodeOutputCmdBaseMixin, TestCase):
    command = "utilization"
    filter_name = "cpu"
    filter_node_and_name = "ram"
    label = "Utilization"
    set_id = "-utilization"

    def setUp(self):
        self.lib = mock.Mock(spec_set=["node", "cluster_property"])
        self.node = mock.Mock(spec_set=["get_config_dto"])
        self.lib.node = self.node
        self.cluster_property = mock.Mock(
            spec_set=["get_properties", "get_properties_metadata"]
        )
        self.lib.cluster_property = self.cluster_property
        self._set_config()

    def _call_cmd(self, argv, modifiers=None):
        node_utilization_output_cmd(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def _set_config(
        self,
        node_config: Optional[CibNodeListDto] = None,
        properties_config: Optional[ListCibNvsetDto] = None,
    ):
        self.node.get_config_dto.return_value = FIXTURE_NODE_CONFIG
        self.cluster_property.get_properties.return_value = (
            FIXTURE_PLACEMENT_STRATEGY_SET
        )
        self.cluster_property.get_properties_metadata.return_value = (
            ClusterPropertyMetadataDto(
                properties_metadata=[], readonly_properties=[]
            )
        )
        if node_config:
            self.node.get_config_dto.return_value = node_config
        if properties_config:
            self.cluster_property.get_properties.return_value = (
                properties_config
            )

    def _set_empty_config(self):
        self.node.get_config_dto.return_value = CibNodeListDto(nodes=[])
        self.cluster_property.get_properties.return_value = (
            FIXTURE_PLACEMENT_STRATEGY_SET
        )
        self.cluster_property.get_properties_metadata.return_value = (
            ClusterPropertyMetadataDto(
                properties_metadata=[], readonly_properties=[]
            )
        )

    @mock.patch("pcs.utils.reports_output.warn")
    def test_utilization_warning(self, mock_warn, mock_print):
        self._set_empty_config()
        self.cluster_property.get_properties.return_value = ListCibNvsetDto(
            nvsets=[]
        )
        self._call_cmd([])
        self._assert_lib_mocks(utilization_warning=True)
        mock_print.assert_not_called()
        mock_warn.assert_called_once_with(UTILIZATION_WARNING)

    def test_force_option_not_supported(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"force": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--force' is not supported in this command",
        )
        mock_print.assert_not_called()



import os
from unittest import TestCase, mock

from pcs.cli import client
from pcs.cli.common.errors import CmdLineInputError
from pcs.common import reports
from pcs.common.file_type_codes import PCS_KNOWN_HOSTS
from pcs.common.host import Destination, PcsKnownHost
from pcs.lib.host.config.exporter import Exporter as KnownHostsExporter
from pcs.lib.host.config.types import KnownHosts

from pcs_test.tools import fixture
from pcs_test.tools.command_env.calls import CallListBuilder, Queue
from pcs_test.tools.command_env.mock_node_communicator import (
    NodeCommunicator,
    place_multinode_call,
)
from pcs_test.tools.command_env.mock_raw_file import (
    RawFileExistsCall,
    RawFileReadCall,
    RawFileWriteCall,
    get_raw_file_mock,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.misc import dict_to_modifiers


def fixture_known_hosts_file_content(
    data_version, hosts: dict[str, PcsKnownHost]
) -> str:
    return KnownHostsExporter.export(
        KnownHosts(
            format_version=1, data_version=data_version, known_hosts=hosts
        )
    ).decode("utf-8")


class LocalAuthCmd(TestCase):
    def setUp(self):
        self.path = os.path.expanduser("~/.pcs/known-hosts")

        self.mock_print_to_stderr = mock.patch(
            "pcs.cli.reports.output.print_to_stderr"
        ).start()
        self.mock_geteuid = mock.patch(
            "pcs.cli.client.os.geteuid", return_value=123
        ).start()
        self.mock_get_user_and_pass = mock.patch(
            "pcs.utils.get_user_and_pass",
            return_value=("hacluster", "password"),
        ).start()
        self.mock_report_processor = (
            mock.patch(
                "pcs.cli.client.ReportProcessorToConsole",
                return_value=MockLibraryReportProcessor(),
            )
            .start()
            .return_value
        )
        self.mock_mkdir = mock.patch("pcs.cli.client.os.mkdir").start()
        self.mock_chmod = mock.patch("pcs.cli.client.os.chmod").start()

        self.addCleanup(mock.patch.stopall)

    def fixture_auth_call(self, successful: bool = True):
        calls = CallListBuilder()
        place_multinode_call(
            calls,
            action="remote/auth",
            node_labels=["localhost"],
            output="TOKEN" if successful else "",
            param_list=[("username", "hacluster"), ("password", "password")],
            name="auth",
        )

        mock.patch(
            "pcs.cli.client._get_node_communicator",
            return_value=NodeCommunicator(Queue(calls)),
        ).start()

    def test_superuser(self):
        self.mock_geteuid.return_value = 0

        with self.assertRaises(SystemExit) as cm:
            client.local_auth_cmd(None, [], dict_to_modifiers({}))

        self.assertEqual(cm.exception.code, 1)
        self.mock_print_to_stderr.assert_called_once()
        self.mock_get_user_and_pass.assert_not_called()
        self.mock_mkdir.assert_not_called()
        self.mock_chmod.assert_not_called()

    def test_too_many_args(self):
        with self.assertRaises(CmdLineInputError):
            client.local_auth_cmd(None, [123, 456], dict_to_modifiers({}))
        self.mock_print_to_stderr.assert_not_called()
        self.mock_get_user_and_pass.assert_not_called()
        self.mock_mkdir.assert_not_called()
        self.mock_chmod.assert_not_called()

    def test_arg_not_a_port_number(self):
        with self.assertRaises(CmdLineInputError):
            client.local_auth_cmd(None, ["abc"], dict_to_modifiers({}))
        self.mock_print_to_stderr.assert_not_called()
        self.mock_get_user_and_pass.assert_not_called()
        self.mock_mkdir.assert_not_called()
        self.mock_chmod.assert_not_called()

    def test_bad_credentials(self):
        self.fixture_auth_call(False)

        with self.assertRaises(SystemExit) as cm:
            client.local_auth_cmd(None, [], dict_to_modifiers({}))
        self.assertEqual(cm.exception.code, 1)

        self.mock_report_processor.assert_reports(
            [
                fixture.error(
                    reports.codes.INCORRECT_CREDENTIALS,
                    context=reports.dto.ReportItemContextDto("localhost"),
                )
            ]
        )
        self.mock_get_user_and_pass.assert_called_once_with()
        self.mock_print_to_stderr.assert_not_called()
        self.mock_mkdir.assert_not_called()
        self.mock_chmod.assert_not_called()

    def test_success(self):
        self.fixture_auth_call()
        calls = CallListBuilder()
        calls.place(
            "exists",
            RawFileExistsCall(file_type_code=PCS_KNOWN_HOSTS, path=self.path),
        )
        calls.place(
            "read",
            RawFileReadCall(
                PCS_KNOWN_HOSTS,
                self.path,
                content=fixture_known_hosts_file_content(1, {}),
            ),
        )
        calls.place(
            "write",
            RawFileWriteCall(
                PCS_KNOWN_HOSTS,
                self.path,
                file_data=fixture_known_hosts_file_content(
                    1,
                    {
                        "localhost": PcsKnownHost(
                            "localhost",
                            "TOKEN",
                            [Destination("localhost", 2224)],
                        )
                    },
                ).encode("utf-8"),
                can_overwrite=True,
            ),
        )

        mock.patch(
            "pcs.cli.client.RawFile", new=get_raw_file_mock(Queue(calls))
        ).start()

        client.local_auth_cmd(None, [], dict_to_modifiers({}))

        self.mock_report_processor.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("localhost"),
                )
            ]
        )
        self.mock_print_to_stderr.assert_not_called()
        self.mock_get_user_and_pass.assert_called_once_with()
        self.mock_mkdir.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )
        self.mock_chmod.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )

    def test_file_no_existed(self):
        self.fixture_auth_call()
        calls = CallListBuilder()
        calls.place(
            "exists",
            RawFileExistsCall(
                file_type_code=PCS_KNOWN_HOSTS, path=self.path, exists=False
            ),
        )
        calls.place(
            "write",
            RawFileWriteCall(
                PCS_KNOWN_HOSTS,
                self.path,
                file_data=fixture_known_hosts_file_content(
                    1,
                    {
                        "localhost": PcsKnownHost(
                            "localhost",
                            "TOKEN",
                            [Destination("localhost", 2224)],
                        )
                    },
                ).encode("utf-8"),
                can_overwrite=True,
            ),
        )

        mock.patch(
            "pcs.cli.client.RawFile", new=get_raw_file_mock(Queue(calls))
        ).start()

        client.local_auth_cmd(None, [], dict_to_modifiers({}))

        self.mock_report_processor.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("localhost"),
                )
            ]
        )
        self.mock_print_to_stderr.assert_not_called()
        self.mock_get_user_and_pass.assert_called_once_with()
        self.mock_mkdir.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )
        self.mock_chmod.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )

    def test_file_read_error(self):
        self.fixture_auth_call()
        calls = CallListBuilder()
        calls.place(
            "exists",
            RawFileExistsCall(file_type_code=PCS_KNOWN_HOSTS, path=self.path),
        )
        calls.place(
            "read",
            RawFileReadCall(PCS_KNOWN_HOSTS, self.path, exception_msg="Error"),
        )

        mock.patch(
            "pcs.cli.client.RawFile", new=get_raw_file_mock(Queue(calls))
        ).start()

        with self.assertRaises(SystemExit) as cm:
            client.local_auth_cmd(None, [], dict_to_modifiers({}))
        self.assertEqual(cm.exception.code, 1)

        self.mock_report_processor.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("localhost"),
                ),
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="read",
                    reason="Error",
                    file_path=self.path,
                ),
            ]
        )
        self.mock_print_to_stderr.assert_not_called()
        self.mock_get_user_and_pass.assert_called_once_with()
        self.mock_mkdir.assert_not_called()
        self.mock_chmod.assert_not_called()

    def test_file_write_error(self):
        self.fixture_auth_call()
        calls = CallListBuilder()
        calls.place(
            "exists",
            RawFileExistsCall(file_type_code=PCS_KNOWN_HOSTS, path=self.path),
        )
        calls.place(
            "read",
            RawFileReadCall(
                PCS_KNOWN_HOSTS,
                self.path,
                content=fixture_known_hosts_file_content(1, {}),
            ),
        )
        calls.place(
            "write",
            RawFileWriteCall(
                PCS_KNOWN_HOSTS,
                self.path,
                file_data=fixture_known_hosts_file_content(
                    1,
                    {
                        "localhost": PcsKnownHost(
                            "localhost",
                            "TOKEN",
                            [Destination("localhost", 2224)],
                        )
                    },
                ).encode("utf-8"),
                can_overwrite=True,
                exception_msg="Error",
            ),
        )

        mock.patch(
            "pcs.cli.client.RawFile", new=get_raw_file_mock(Queue(calls))
        ).start()

        with self.assertRaises(SystemExit) as cm:
            client.local_auth_cmd(None, [], dict_to_modifiers({}))
        self.assertEqual(cm.exception.code, 1)

        self.mock_report_processor.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("localhost"),
                ),
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="write",
                    reason="Error",
                    file_path=self.path,
                ),
            ]
        )
        self.mock_print_to_stderr.assert_not_called()
        self.mock_get_user_and_pass.assert_called_once_with()
        self.mock_mkdir.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )
        self.mock_chmod.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )

    def test_cannot_create_folder(self):
        self.mock_mkdir.side_effect = FileNotFoundError()
        self.fixture_auth_call()
        calls = CallListBuilder()
        calls.place(
            "exists",
            RawFileExistsCall(file_type_code=PCS_KNOWN_HOSTS, path=self.path),
        )
        calls.place(
            "read",
            RawFileReadCall(
                PCS_KNOWN_HOSTS,
                self.path,
                content=fixture_known_hosts_file_content(1, {}),
            ),
        )

        mock.patch(
            "pcs.cli.client.RawFile", new=get_raw_file_mock(Queue(calls))
        ).start()

        with self.assertRaises(SystemExit) as cm:
            client.local_auth_cmd(None, [], dict_to_modifiers({}))

        self.mock_report_processor.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("localhost"),
                )
            ]
        )
        self.assertEqual(cm.exception.code, 1)
        self.mock_print_to_stderr.assert_called_once_with(
            "Error: FileNotFoundError"
        )
        self.mock_get_user_and_pass.assert_called_once_with()
        self.mock_mkdir.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )
        self.mock_chmod.assert_not_called()

    def test_folder_already_exists(self):
        self.mock_mkdir.side_effect = FileExistsError()
        self.fixture_auth_call()
        calls = CallListBuilder()
        calls.place(
            "exists",
            RawFileExistsCall(file_type_code=PCS_KNOWN_HOSTS, path=self.path),
        )
        calls.place(
            "read",
            RawFileReadCall(
                PCS_KNOWN_HOSTS,
                self.path,
                content=fixture_known_hosts_file_content(1, {}),
            ),
        )
        calls.place(
            "write",
            RawFileWriteCall(
                PCS_KNOWN_HOSTS,
                self.path,
                file_data=fixture_known_hosts_file_content(
                    1,
                    {
                        "localhost": PcsKnownHost(
                            "localhost",
                            "TOKEN",
                            [Destination("localhost", 2224)],
                        )
                    },
                ).encode("utf-8"),
                can_overwrite=True,
            ),
        )

        mock.patch(
            "pcs.cli.client.RawFile", new=get_raw_file_mock(Queue(calls))
        ).start()

        client.local_auth_cmd(None, [], dict_to_modifiers({}))

        self.mock_report_processor.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("localhost"),
                )
            ]
        )
        self.mock_print_to_stderr.assert_not_called()
        self.mock_get_user_and_pass.assert_called_once_with()
        self.mock_mkdir.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )
        self.mock_chmod.assert_called_once_with(
            os.path.expanduser("~/.pcs"), 0o700
        )



from unittest import (
    TestCase,
)

from pcs.common import (
    reports,
)
from pcs.lib.commands import resource
from pcs.lib.resource_agent.const import PRIMITIVE_META, STONITH_META
from pcs.lib.resource_agent.types import ResourceAgentName

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_raise_library_error
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.metadata_dto import (
    FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
    FIXTURE_KNOWN_META_NAMES_STONITH_META,
)

_AGENT_NAME_PCMK_DUMMY = ResourceAgentName("ocf", "pacemaker", "Dummy")
_AGENT_NAME_STONITH = ResourceAgentName("stonith", None, "Dummy")
_AGENT_NAME_PCMK_STATEFUL = ResourceAgentName("ocf", "pacemaker", "Stateful")


def fixture_primitive(
    resource_agent=_AGENT_NAME_PCMK_DUMMY,
    is_cloned=False,
    is_grouped=False,
    meta_nvpairs_xml="",
    primitive_inner="",
):
    meta_attributes_xml = meta_attributes_xml_clone = ""
    if meta_nvpairs_xml:
        if is_cloned:
            meta_attributes_xml_clone = f"""
                <meta_attributes id="A-clone-meta_attributes">
                    {meta_nvpairs_xml}
                </meta_attributes>
            """
        else:
            meta_attributes_xml = f"""
                <meta_attributes id="A-meta_attributes">
                    {meta_nvpairs_xml}
                </meta_attributes>
            """

    clone_start = clone_end = ""
    if is_cloned:
        clone_start = """<clone id="A-clone">"""
        clone_end = "</clone>"

    group_start = group_end = ""
    if is_grouped:
        group_start = """<group id="G">"""
        group_end = "</group>"

    return f"""
        <resources>
            {clone_start}
                {group_start}
                    <primitive class="{resource_agent.standard}" id="A" 
                        provider="{resource_agent.provider}"
                        type="{resource_agent.type}"
                    >
                        {meta_attributes_xml}
                        {primitive_inner}
                    </primitive>
                {group_end}
                {meta_attributes_xml_clone}
            {clone_end}
        </resources>
    """


def fixture_resource_meta_stateful(
    use_legacy_roles=False,
    meta_nvpairs="",
    clone_inner="",
    is_grouped=False,
    is_promotable=True,
):
    clone_el_tag = "clone"
    role_promoted = "Promoted"
    role_unpromoted = "Unpromoted"

    if use_legacy_roles:
        clone_el_tag = "master" if is_promotable else "clone"
        role_promoted = "Master"
        role_unpromoted = "Slave"
    elif is_promotable:
        meta_nvpairs += """
            <nvpair id="A-clone-meta_attributes-promotable" name="promotable"
                value="true"
            />
        """

    meta_attributes_xml = ""
    if meta_nvpairs:
        meta_attributes_xml = f"""
            <meta_attributes id="A-clone-meta_attributes">
                {meta_nvpairs}
            </meta_attributes>
        """

    group_start = group_end = ""
    if is_grouped:
        group_start = """<group id="G">"""
        group_end = "</group>"

    return f"""
        <resources>
            <{clone_el_tag} id="A-clone">
                {group_start}
                    <primitive id="A" class="ocf" type="Stateful"
                        provider="pacemaker"
                    >
                        <operations>
                            <op name="demote" interval="0s" timeout="10s"
                                id="A-demote-interval-0s"
                            />
                            <op name="monitor" interval="10s" timeout="20s"
                                role="{role_promoted}" id="A-monitor-interval-10s"
                            />
                            <op name="monitor" interval="11s" timeout="20s"
                                role="{role_unpromoted}" id="A-monitor-interval-11s"
                            />
                            <op name="notify" interval="0s" timeout="5s"
                                id="A-notify-interval-0s"
                            />
                            <op name="promote" interval="0s" timeout="10s"
                                id="A-promote-interval-0s"
                            />
                            <op name="reload-agent" interval="0s" timeout="10s"
                                id="A-reload-agent-interval-0s"
                            />
                            <op name="start" interval="0s" timeout="20s"
                                id="A-start-interval-0s"
                            />
                            <op name="stop" interval="0s" timeout="20s"
                                id="A-stop-interval-0s"
                            />
                        </operations>
                    </primitive>
                {group_end}
                {meta_attributes_xml}
                {clone_inner}
            </{clone_el_tag}>
        </resources>
    """


class UpdateMeta(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_update_nonexistent_resource(self):
        self.config.runner.cib.load(filename="cib-all.xml")
        self.env_assist.assert_raise_library_error(
            lambda: resource.update_meta(
                self.env_assist.get_env(), "Rx", {"priority": "1"}, []
            ),
            reports=[
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="Rx",
                    expected_types=["resource"],
                    context_type="",
                    context_id="",
                ),
            ],
            expected_in_processor=False,
        )

    def test_wrong_id_type_found(self):
        self.config.runner.cib.load(
            tags="""
            <tags>
              <tag id="all-vms">
                <obj_ref id="vm1"/>
                <obj_ref id="vm2"/>
              </tag>
            </tags>
                """
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.update_meta(
                self.env_assist.get_env(), "all-vms", {"priority": "1"}, []
            ),
            reports=[
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="all-vms",
                    expected_types=["resource"],
                    current_type="tag",
                ),
            ],
            expected_in_processor=False,
        )

    def test_meta_attr_elem_missing(self):
        self.config.runner.cib.load(resources=fixture_primitive())
        self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-priority" name="priority"
                        value="1"
                    />
                """
            )
        )
        resource.update_meta(
            self.env_assist.get_env(), "A", {"priority": "1"}, []
        )

    def test_no_new_attrs_element_not_added(self):
        self.config.runner.cib.load(resources=fixture_primitive())
        self.config.corosync_conf.load()
        self.config.env.push_cib(resources=fixture_primitive())
        resource.update_meta(
            self.env_assist.get_env(), "A", {"priority": ""}, []
        )

    def test_existing_meta_attr_elem_add_remove(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-resource-stickiness"
                        name="resource-stickiness" value="0"
                    />
                    <nvpair id="A-meta_attributes-failure-timeout"
                    name="failure-timeout" value="10s"
                    />
                """,
            )
        )
        self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-resource-stickiness"
                        name="resource-stickiness" value="0"
                    />
                    <nvpair id="A-meta_attributes-priority" name="priority"
                        value="1"
                    />
                """,
            )
        )
        resource.update_meta(
            self.env_assist.get_env(),
            "A",
            {"priority": "1", "failure-timeout": ""},
            [],
        )

    def test_conflict_existing_node_name(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(),
        )
        self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load(node_name_list=["node1"])
        self.env_assist.assert_raise_library_error(
            lambda: resource.update_meta(
                self.env_assist.get_env(),
                "A",
                {"remote-node": "node1"},
                [],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.GUEST_NODE_NAME_ALREADY_EXISTS,
                    node_name="node1",
                ),
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_ADD_GUEST,
                    reports.codes.FORCE,
                ),
            ]
        )

    def test_update_guest_attr_protected_force(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(),
        )
        self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-remote-node"
                        name="remote-node" value="node1"
                    />
                """
            )
        )
        resource.update_meta(
            self.env_assist.get_env(),
            "A",
            {"remote-node": "node1"},
            [reports.codes.FORCE],
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(reports.codes.USE_COMMAND_NODE_ADD_GUEST),
            ]
        )

    def test_update_legacy_clone_element_tag_and_reset_promotable(self):
        self.config.runner.cib.load(
            resources=fixture_resource_meta_stateful(
                use_legacy_roles=True,
            )
        )
        self.config.corosync_conf.load()
        agent = ResourceAgentName("ocf", "pacemaker", "Stateful")
        self.config.runner.pcmk.load_agent(agent_name=agent.full_name)
        self.config.env.push_cib(
            resources=fixture_resource_meta_stateful(
                meta_nvpairs="""
                    <nvpair id="A-clone-meta_attributes-priority"
                        name="priority" value="1"
                    />
                """,
                is_promotable=False,
            )
        )
        resource.update_meta(
            self.env_assist.get_env(),
            "A-clone",
            {"priority": "1", "promotable": ""},
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
            ]
        )


class UpdateMetaCheckCloneIncompatibleMetaAttrs(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success_cloned_primitive_priority(self):
        self.config.runner.cib.load(resources=fixture_primitive(is_cloned=True))
        self.config.corosync_conf.load()
        self.config.runner.pcmk.load_agent(
            agent_name=_AGENT_NAME_PCMK_DUMMY.full_name
        )
        self.config.env.push_cib(
            resources=fixture_primitive(
                is_cloned=True,
                meta_nvpairs_xml="""
                    <nvpair id="A-clone-meta_attributes-priority"
                        name="priority" value="1"
                    />
                """,
            )
        )
        resource.update_meta(
            self.env_assist.get_env(), "A-clone", {"priority": "1"}, []
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
            ]
        )

    def test_success_cloned_primitive_promotable(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(
                resource_agent=_AGENT_NAME_PCMK_STATEFUL,
                is_cloned=True,
            )
        )
        self.config.corosync_conf.load()
        self.config.runner.pcmk.load_agent(
            agent_name=_AGENT_NAME_PCMK_STATEFUL.full_name
        )
        self.config.env.push_cib(
            resources=fixture_primitive(
                resource_agent=_AGENT_NAME_PCMK_STATEFUL,
                is_cloned=True,
                meta_nvpairs_xml="""
                    <nvpair id="A-clone-meta_attributes-promotable" 
                        name="promotable" value="true"
                    />
                """,
            )
        )
        resource.update_meta(
            self.env_assist.get_env(), "A-clone", {"promotable": "true"}, []
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
            ]
        )

    def test_success_cloned_primitive_promotable_ocf_old(self):
        # Despite the test name, agent older than OCF 1.1 can be promoted
        # because the old standard doesn't allow for promotability checks
        agent_name = ResourceAgentName("ocf", "heartbeat", "Dummy")
        self.config.runner.cib.load(
            resources=fixture_primitive(
                resource_agent=agent_name,
                is_cloned=True,
            )
        )
        self.config.corosync_conf.load()
        self.config.runner.pcmk.load_agent(agent_name=agent_name.full_name)
        self.config.env.push_cib(
            resources=fixture_primitive(
                resource_agent=agent_name,
                is_cloned=True,
                meta_nvpairs_xml="""
                    <nvpair id="A-clone-meta_attributes-promotable" 
                        name="promotable" value="true"
                    />
                """,
            )
        )
        resource.update_meta(
            self.env_assist.get_env(), "A-clone", {"promotable": "true"}, []
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
            ]
        )

    def test_fail_cloned_primitive_promotable_incompatible(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(is_cloned=True),
        )
        self.config.corosync_conf.load()
        self.config.runner.pcmk.load_agent(
            agent_name=_AGENT_NAME_PCMK_DUMMY.full_name
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.update_meta(
                self.env_assist.get_env(),
                "A-clone",
                {"promotable": "true"},
                [],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
                fixture.error(
                    reports.codes.RESOURCE_CLONE_INCOMPATIBLE_META_ATTRIBUTES,
                    attribute="promotable",
                    resource_agent=_AGENT_NAME_PCMK_DUMMY.to_dto(),
                    resource_id="A",
                    group_id=None,
                    force_code=reports.codes.FORCE,
                ),
            ]
        )

    def test_fail_cloned_primitive_promotable_incompatible_force(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(is_cloned=True),
        )
        self.config.corosync_conf.load()
        self.config.runner.pcmk.load_agent(
            agent_name=_AGENT_NAME_PCMK_DUMMY.full_name
        )
        self.config.env.push_cib(
            resources=fixture_primitive(
                is_cloned=True,
                meta_nvpairs_xml="""
                    <nvpair id="A-clone-meta_attributes-promotable"
                        name="promotable" value="true"
                    />
                """,
            )
        )
        resource.update_meta(
            self.env_assist.get_env(),
            "A-clone",
            {"promotable": "true"},
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
                fixture.warn(
                    reports.codes.RESOURCE_CLONE_INCOMPATIBLE_META_ATTRIBUTES,
                    attribute="promotable",
                    resource_agent=_AGENT_NAME_PCMK_DUMMY.to_dto(),
                    resource_id="A",
                    group_id=None,
                ),
            ]
        )

    def _subtest_fail_cloned_primitive_non_ocf_incompatible_attrs(self, attr):
        # Cannot use real subtest because of env_assist call stack, it needs
        # initialization and teardown between test runs
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <clone id="non-ocf-clone">
                        <primitive id="non-ocf" class="systemd"
                          type="pcsmock" />
                    </clone>
                </resources>
            """,
        )
        self.config.corosync_conf.load()
        agent = ResourceAgentName("systemd", None, "pcsmock")
        self.env_assist.assert_raise_library_error(
            lambda: resource.update_meta(
                self.env_assist.get_env(),
                "non-ocf-clone",
                {attr: "true"},  # noqa: B023
                [],
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
                fixture.error(
                    reports.codes.RESOURCE_CLONE_INCOMPATIBLE_META_ATTRIBUTES,
                    attribute=attr,
                    resource_agent=agent.to_dto(),
                    resource_id="non-ocf",
                    group_id=None,
                ),
            ]
        )

    def test_fail_cloned_primitive_non_ocf_promotable(self):
        self._subtest_fail_cloned_primitive_non_ocf_incompatible_attrs(
            "promotable"
        )

    def test_fail_cloned_primitive_non_ocf_globally_unique(self):
        self._subtest_fail_cloned_primitive_non_ocf_incompatible_attrs(
            "globally-unique"
        )

    def test_fail_cloned_group_promotable(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(is_cloned=True, is_grouped=True),
        )
        self.config.corosync_conf.load()
        self.config.runner.pcmk.load_agent(
            agent_name=_AGENT_NAME_PCMK_DUMMY.full_name
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.update_meta(
                self.env_assist.get_env(),
                "A-clone",
                {"promotable": "true"},
                [],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
                fixture.error(
                    reports.codes.RESOURCE_CLONE_INCOMPATIBLE_META_ATTRIBUTES,
                    attribute="promotable",
                    resource_agent=_AGENT_NAME_PCMK_DUMMY.to_dto(),
                    resource_id="A",
                    group_id="G",
                    force_code=reports.codes.FORCE,
                ),
            ]
        )

    def test_fail_cloned_group_promotable_force(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(is_cloned=True, is_grouped=True),
        )
        self.config.corosync_conf.load()
        self.config.runner.pcmk.load_agent(
            agent_name=_AGENT_NAME_PCMK_DUMMY.full_name
        )
        self.config.env.push_cib(
            resources=fixture_primitive(
                is_cloned=True,
                is_grouped=True,
                meta_nvpairs_xml="""
                    <nvpair id="A-clone-meta_attributes-promotable"
                        name="promotable" value="true"
                    />
                """,
            )
        )
        resource.update_meta(
            self.env_assist.get_env(),
            "A-clone",
            {"promotable": "true"},
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_UNSUPPORTED_TYPE,
                    meta_type_list=["bundle", "clone"],
                ),
                fixture.warn(
                    reports.codes.RESOURCE_CLONE_INCOMPATIBLE_META_ATTRIBUTES,
                    attribute="promotable",
                    resource_agent=_AGENT_NAME_PCMK_DUMMY.to_dto(),
                    resource_id="A",
                    group_id="G",
                ),
            ]
        )


class UpdateMetaRemoveUpdatedGuestNode(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def _update_guest_attr_success(
        self,
        old_node_addr="",
        new_node_addr="",
        is_nvset_empty=False,
        is_node_removed=False,
        is_forced=False,
        report_list=None,
    ):
        self.config.runner.cib.load(
            resources=fixture_primitive(
                meta_nvpairs_xml=(
                    f"""
                    <nvpair id="A-meta_attributes-remote-node"
                        name="remote-node" value="{old_node_addr}"
                    />
                """
                    if old_node_addr
                    else ""
                ),
            )
        )
        # we do not want to read the metadata when we are not adding any new
        # attributes
        if new_node_addr:
            self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                # For removeal, we want to hack the fixture to get the meta
                # attributes element without nvpairs
                meta_nvpairs_xml=(
                    " "
                    if is_nvset_empty
                    else f"""
                    <nvpair id="A-meta_attributes-remote-node"
                        name="remote-node" value="{new_node_addr}"
                    />
                """
                )
            )
        )
        if is_node_removed:
            self.config.runner.pcmk.remove_node(old_node_addr)
        env = self.env_assist.get_env()
        resource.update_meta(
            env,
            "A",
            {"remote-node": new_node_addr},
            [reports.codes.FORCE] if is_forced else [],
        )
        self.env_assist.assert_reports(report_list if report_list else [])

    def _update_guest_attr_fail(
        self, report_list, old_node_addr="", new_node_addr="", is_forced=False
    ):
        self.config.runner.cib.load(
            resources=fixture_primitive(
                meta_nvpairs_xml=(
                    f"""
                    <nvpair id="A-meta_attributes-remote-node"
                        name="remote-node" value="{old_node_addr}"
                    />
                """
                    if old_node_addr
                    else ""
                ),
            )
        )
        if new_node_addr:
            self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        env = self.env_assist.get_env()
        assert_raise_library_error(
            lambda: resource.update_meta(
                env,
                "A",
                {"remote-node": new_node_addr},
                [reports.codes.FORCE] if is_forced else [],
            )
        )
        self.env_assist.assert_reports(report_list)

    def test_not_called_add_remote_node(self):
        self._update_guest_attr_fail(
            new_node_addr="rnode",
            report_list=[
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_ADD_GUEST,
                    reports.codes.FORCE,
                ),
            ],
        )

    def test_not_called_add_remote_node_force(self):
        self._update_guest_attr_success(
            new_node_addr="rnode",
            is_forced=True,
            report_list=[
                fixture.warn(reports.codes.USE_COMMAND_NODE_ADD_GUEST),
            ],
        )

    def test_not_called_change_remote_node(self):
        self._update_guest_attr_fail(
            old_node_addr="rnode2",
            new_node_addr="rnode",
            report_list=[
                fixture.error(
                    reports.codes.USE_COMMAND_REMOVE_AND_ADD_GUEST_NODE,
                    reports.codes.FORCE,
                ),
            ],
        )

    def test_called_change_remote_node_force(self):
        self._update_guest_attr_success(
            old_node_addr="rnode2",
            new_node_addr="rnode",
            is_forced=True,
            is_node_removed=True,
            report_list=[
                fixture.warn(
                    reports.codes.USE_COMMAND_REMOVE_AND_ADD_GUEST_NODE
                ),
            ],
        )

    def test_not_called_fake_change_remote_node(self):
        self._update_guest_attr_fail(
            old_node_addr="rnode",
            new_node_addr="rnode",
            report_list=[
                fixture.error(
                    reports.codes.GUEST_NODE_NAME_ALREADY_EXISTS,
                    node_name="rnode",
                )
            ],
        )

    def test_not_called_fake_change_remote_node_force(self):
        self._update_guest_attr_fail(
            old_node_addr="rnode",
            new_node_addr="rnode",
            is_forced=True,
            report_list=[
                fixture.error(
                    reports.codes.GUEST_NODE_NAME_ALREADY_EXISTS,
                    node_name="rnode",
                )
            ],
        )

    def test_not_called_remove_remote_node(self):
        self._update_guest_attr_fail(
            old_node_addr="rnode2",
            new_node_addr="",
            report_list=[
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_GUEST,
                    resource_id=None,
                    force_code=reports.codes.FORCE,
                ),
            ],
        )

    def test_called_remove_remote_node_force(self):
        self._update_guest_attr_success(
            old_node_addr="rnode2",
            new_node_addr="",
            is_nvset_empty=True,
            is_node_removed=True,
            is_forced=True,
            report_list=[
                fixture.warn(
                    reports.codes.USE_COMMAND_NODE_REMOVE_GUEST,
                    resource_id=None,
                ),
            ],
        )


class UpdateMetaValidateMetaAttributes(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive_invalid_meta_attribute_name(self):
        self.config.runner.cib.load(resources=fixture_primitive())
        self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-name"
                        name="name" value="value"
                    />
                """,
            )
        )

        resource.update_meta(
            self.env_assist.get_env(), "A", {"name": "value"}, []
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_UNKNOWN_TO_PCMK,
                    unknown_meta=["name"],
                    known_meta=FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
                    meta_types=[PRIMITIVE_META],
                )
            ]
        )

    def test_primitive_invalid_stonith_meta_attribute_name(self):
        self.config.runner.cib.load(resources=fixture_primitive())
        self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-provides"
                        name="provides" value="everything"
                    />
                """,
            )
        )

        resource.update_meta(
            self.env_assist.get_env(), "A", {"provides": "everything"}, []
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_UNKNOWN_TO_PCMK,
                    unknown_meta=["provides"],
                    known_meta=FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
                    meta_types=[PRIMITIVE_META],
                )
            ]
        )

    def test_group_invalid_stonith_meta_attribute_name(self):
        self.config.runner.cib.load(resources=fixture_primitive())
        self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-provides"
                        name="provides" value="everything"
                    />
                """,
            )
        )

        resource.update_meta(
            self.env_assist.get_env(), "A", {"provides": "everything"}, []
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_UNKNOWN_TO_PCMK,
                    unknown_meta=["provides"],
                    known_meta=FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
                    meta_types=[PRIMITIVE_META],
                )
            ]
        )

    def test_stonith_invalid_meta_attribute_name(self):
        self.config.runner.cib.load(
            resources=fixture_primitive(resource_agent=_AGENT_NAME_STONITH)
        )
        self.config.runner.pcmk.load_crm_resource_metadata()
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                resource_agent=_AGENT_NAME_STONITH,
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-name"
                        name="name" value="value"
                    />
                """,
            )
        )

        resource.update_meta(
            self.env_assist.get_env(), "A", {"name": "value"}, []
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.META_ATTRS_UNKNOWN_TO_PCMK,
                    unknown_meta=["name"],
                    known_meta=FIXTURE_KNOWN_META_NAMES_STONITH_META,
                    meta_types=[STONITH_META],
                )
            ]
        )

    def test_primitive_error_loading_metadata(self):
        self.config.runner.cib.load(resources=fixture_primitive())
        self.config.runner.pcmk.load_crm_resource_metadata(
            stdout="output", returncode=1
        )
        self.config.corosync_conf.load()
        self.config.env.push_cib(
            resources=fixture_primitive(
                meta_nvpairs_xml="""
                    <nvpair id="A-meta_attributes-name"
                        name="name" value="value"
                    />
                """,
            )
        )

        resource.update_meta(
            self.env_assist.get_env(), "A", {"name": "value"}, []
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent="primitive-meta",
                    reason="output",
                ),
                fixture.warn(
                    reports.codes.META_ATTRS_NOT_VALIDATED_LOADING_ERROR
                ),
            ]
        )



from unittest import TestCase

from pcs.common.host import Destination, PcsKnownHost
from pcs.lib.host.config.facade import Facade as KnownHostsFacade
from pcs.lib.host.config.types import KnownHosts


class Facade(TestCase):
    def setUp(self):
        self.facade = KnownHostsFacade(
            KnownHosts(
                format_version=1,
                data_version=10,
                known_hosts={
                    "a": PcsKnownHost(
                        name="a",
                        token="abcd",
                        dest_list=[Destination("10.0.0.1", 2224)],
                    )
                },
            )
        )

    def test_create(self):
        facade = KnownHostsFacade.create()

        self.assertEqual(1, facade.data_version)
        self.assertEqual(dict(), facade.known_hosts)

    def test_ok(self):
        self.assertEqual(10, self.facade.data_version)
        self.assertEqual(
            {
                "a": PcsKnownHost(
                    name="a",
                    token="abcd",
                    dest_list=[Destination("10.0.0.1", 2224)],
                )
            },
            self.facade.known_hosts,
        )

    def test_update_known_hosts_add_new_host(self):
        self.facade.update_known_hosts(
            [
                PcsKnownHost(
                    name="b",
                    token="wxyz",
                    dest_list=[Destination("10.0.0.2", 2224)],
                )
            ]
        )

        self.assertEqual(
            {
                "a": PcsKnownHost(
                    name="a",
                    token="abcd",
                    dest_list=[Destination("10.0.0.1", 2224)],
                ),
                "b": PcsKnownHost(
                    name="b",
                    token="wxyz",
                    dest_list=[Destination("10.0.0.2", 2224)],
                ),
            },
            self.facade.known_hosts,
        )

    def test_update_known_hosts_rewrite_existing(self):
        self.facade.update_known_hosts(
            [
                PcsKnownHost(
                    name="a",
                    token="wxyz",
                    dest_list=[Destination("10.0.0.2", 2224)],
                )
            ]
        )

        self.assertEqual(
            {
                "a": PcsKnownHost(
                    name="a",
                    token="wxyz",
                    dest_list=[Destination("10.0.0.2", 2224)],
                ),
            },
            self.facade.known_hosts,
        )

    def test_remove_known_hosts(self):
        facade = KnownHostsFacade(
            KnownHosts(
                format_version=1,
                data_version=10,
                known_hosts={
                    "a": PcsKnownHost(
                        name="a",
                        token="abcd",
                        dest_list=[Destination("10.0.0.1", 2224)],
                    ),
                    "b": PcsKnownHost(
                        name="b",
                        token="wxyz",
                        dest_list=[Destination("10.0.0.2", 2224)],
                    ),
                },
            )
        )

        facade.remove_known_hosts(["b", "c"])

        self.assertEqual(
            {
                "a": PcsKnownHost(
                    name="a",
                    token="abcd",
                    dest_list=[Destination("10.0.0.1", 2224)],
                )
            },
            self.facade.known_hosts,
        )

    def test_set_version(self):
        self.assertEqual(self.facade.data_version, 10)
        self.facade.set_data_version(1000)
        self.assertEqual(self.facade.data_version, 1000)



import json
from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import (
    format_cmd_list,
    lines_to_str,
)
from pcs.cli.common.parse_args import (
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    Argv,
    InputModifiers,
)
from pcs.cli.stonith.levels import output as levels_output
from pcs.common.interface.dto import to_dict


def config(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --output-format - supported formats: text, cmd, json
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f", output_format_supported=True)
    output_format = modifiers.get_output_format()
    if argv:
        raise CmdLineInputError

    fencing_topology_dto = lib.fencing_topology.get_config_dto()

    if output_format == OUTPUT_FORMAT_VALUE_JSON:
        output = json.dumps(to_dict(fencing_topology_dto))
    elif output_format == OUTPUT_FORMAT_VALUE_CMD:
        output = format_cmd_list(
            levels_output.stonith_level_config_to_cmd(fencing_topology_dto)
        )
    else:
        output = lines_to_str(
            levels_output.stonith_level_config_to_text(fencing_topology_dto)
        )

    if output:
        print(output)



from typing import Iterable, Mapping, cast

from pcs.common import reports
from pcs.common.types import StringIterable
from pcs.lib import validate
from pcs.lib.resource_agent import CrmResourceAgent, ResourceAgentParameter


def validate_meta_attributes(
    meta_attributes_types: Iterable[CrmResourceAgent],
    meta_attributes_metadata: Iterable[ResourceAgentParameter],
    meta_attributes: Mapping[str, str],
) -> reports.ReportItemList:
    """
    Validate meta attributes.

    meta_attributes_types -- resource type of the meta attributes
    meta_attributes_metadata -- meta attribute definitions of a specific
        resource type
    meta_attributes -- meta attributes to be validated
    """
    return validate.PcmkMetaAttributeNamesIn(
        [parameter.name for parameter in meta_attributes_metadata],
        cast(StringIterable, meta_attributes_types),
    ).validate(
        # Allow removing meta attributes unknown to pacemaker while preventing
        # setting them
        {name: value for name, value in meta_attributes.items() if value != ""}
    )



from textwrap import dedent
from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import str_to_etree


class ElementDescription(AssertPcsMixin, TestCase):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_element_description")
        write_file_to_tmpfile(get_test_resource("cib-all.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def assert_description_in_cib(self, xpath, expected_description=None):
        self.temp_cib.seek(0)
        xpath_result = str_to_etree(self.temp_cib.read()).xpath(xpath)
        cib_description = xpath_result[0] if xpath_result else None
        self.assertEqual(cib_description, expected_description)

    def test_usage(self):
        self.assert_pcs_fail(
            ["cib", "element", "description"],
            stderr_start=dedent("""
                Usage: pcs cib <command>
                    element description <element-id>
            """),
        )

    def test_add_description(self):
        # primitive resource without description
        xpath = "//primitive[@id='R6']/@description"
        self.assert_description_in_cib(xpath, None)

        description = "My description"
        self.assert_pcs_success(
            ["cib", "element", "description", "R6", description],
            stdout_full="",
            stderr_full="",
        )

        self.assert_description_in_cib(xpath, description)

    def test_remove_description(self):
        # alert with description
        xpath = "//alert[@id='alert-all']/@description"
        self.assert_description_in_cib(xpath, "alert all options")

        self.assert_pcs_success(
            ["cib", "element", "description", "alert-all", ""],
            stdout_full="",
            stderr_full="",
        )

        self.assert_description_in_cib(xpath, None)

    def test_show_description(self):
        self.assert_pcs_success(
            ["cib", "element", "description", "alert-all"],
            stdout_full="alert all options\n",
            stderr_full="",
        )

    def test_show_description_no_description(self):
        self.assert_pcs_success(
            ["cib", "element", "description", "R6"],
            stdout_full="",
            stderr_full="",
        )



import json
import shlex

from pcs.cli.common.output import lines_to_str
from pcs.cli.common.parse_args import (
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    InputModifiers,
)
from pcs.common.interface import dto
from pcs.common.pacemaker.tag import CibTagListDto
from pcs.common.str_tools import indent


def tags_to_text(tags_dto: CibTagListDto) -> list[str]:
    result = []
    for tag in tags_dto.tags:
        result.append(tag.id)
        result.extend(indent(list(tag.idref_list)))
    return result


def tags_to_cmd(tags_dto: CibTagListDto) -> list[str]:
    return [
        "pcs -- tag create {tag_id} {idref_list}".format(
            tag_id=shlex.quote(tag.id),
            idref_list=" ".join(shlex.quote(idref) for idref in tag.idref_list),
        )
        for tag in tags_dto.tags
    ]


def print_config(tags_dto: CibTagListDto, modifiers: InputModifiers) -> None:
    output_format = modifiers.get_output_format()
    if output_format == OUTPUT_FORMAT_VALUE_JSON:
        print(json.dumps(dto.to_dict(tags_dto), indent=2))
        return

    if output_format == OUTPUT_FORMAT_VALUE_CMD:
        print(";\n".join(tags_to_cmd(tags_dto)))
        return

    result = lines_to_str(tags_to_text(tags_dto))
    if result:
        print(result)



from unittest import TestCase, mock

from pcs.cli.booth import command as booth_cmd
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers
from pcs.common import reports
from pcs.lib.errors import LibraryError

from pcs_test.tools.misc import dict_to_modifiers

UNSTOPPED_RESOURCES_ERROR_REPORT = reports.ReportItem.error(
    reports.messages.CannotRemoveResourcesNotStopped(["R1"])
)
BOOTH_NOT_IN_CIB_ERROR = reports.ReportItem.error(
    reports.messages.BoothNotExistsInCib(name="booth")
)
FORCEABLE_ERROR_REPORT = reports.ReportItem.error(
    reports.messages.BoothNotExistsInCib(name="booth"),
    force_code=reports.codes.FORCE,
)
INFO_REPORT = reports.ReportItem.info(
    reports.messages.CibRemoveDependantElements({})
)


class SetupTest(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["config_setup"])

    def test_lib_call_minimal(self):
        booth_cmd.config_setup(
            self.lib,
            ["sites", "1.1.1.1", "2.2.2.2", "3.3.3.3"],
            dict_to_modifiers({}),
        )
        self.lib.booth.config_setup.assert_called_once_with(
            ["1.1.1.1", "2.2.2.2", "3.3.3.3"],
            [],
            instance_name=None,
            overwrite_existing=False,
        )

    def test_lib_call_full(self):
        booth_cmd.config_setup(
            self.lib,
            ["sites", "1.1.1.1", "2.2.2.2", "arbitrators", "3.3.3.3"],
            dict_to_modifiers(dict(name="my_booth", force=True)),
        )
        self.lib.booth.config_setup.assert_called_once_with(
            ["1.1.1.1", "2.2.2.2"],
            ["3.3.3.3"],
            instance_name="my_booth",
            overwrite_existing=True,
        )


class DestroyTest(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["config_destroy"])

    def test_too_many_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            booth_cmd.config_destroy(self.lib, ["aaa"], dict_to_modifiers({}))
        self.assertIsNone(cm.exception.message)
        self.lib.booth.config_destroy.assert_not_called()

    def test_lib_call_minimal(self):
        booth_cmd.config_destroy(self.lib, [], dict_to_modifiers({}))
        self.lib.booth.config_destroy.assert_called_once_with(
            ignore_config_load_problems=False,
            instance_name=None,
        )

    def test_lib_call_full(self):
        booth_cmd.config_destroy(
            self.lib, [], dict_to_modifiers(dict(name="my_booth", force=True))
        )
        self.lib.booth.config_destroy.assert_called_once_with(
            ignore_config_load_problems=True,
            instance_name="my_booth",
        )


class AddTicketTest(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["config_ticket_add"])

    def test_lib_call_minimal(self):
        booth_cmd.config_ticket_add(
            self.lib, ["ticketA"], dict_to_modifiers({})
        )
        self.lib.booth.config_ticket_add.assert_called_once_with(
            "ticketA",
            {},
            instance_name=None,
            allow_unknown_options=False,
        )

    def test_lib_call_full(self):
        booth_cmd.config_ticket_add(
            self.lib,
            ["ticketA", "a=A", "b=B"],
            dict_to_modifiers(
                {
                    "name": "my_booth",
                    "force": True,
                    "booth-conf": "C",
                    "booth-key": "K",
                }
            ),
        )
        self.lib.booth.config_ticket_add.assert_called_once_with(
            "ticketA",
            {"a": "A", "b": "B"},
            instance_name="my_booth",
            allow_unknown_options=True,
        )


class RemoveTicketTest(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["config_ticket_remove"])

    def test_lib_call_minimal(self):
        booth_cmd.config_ticket_remove(
            self.lib, ["ticketA"], dict_to_modifiers({})
        )
        self.lib.booth.config_ticket_remove.assert_called_once_with(
            "ticketA",
            instance_name=None,
        )

    def test_lib_call_full(self):
        booth_cmd.config_ticket_remove(
            self.lib,
            ["ticketA"],
            dict_to_modifiers(
                {
                    "name": "my_booth",
                    "booth-conf": "C",
                    "booth-key": "K",
                }
            ),
        )
        self.lib.booth.config_ticket_remove.assert_called_once_with(
            "ticketA",
            instance_name="my_booth",
        )


class CreateTest(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["create_in_cluster"])

    def test_lib_call_minimal(self):
        booth_cmd.create_in_cluster(
            self.lib, ["ip", "1.2.3.4"], dict_to_modifiers({})
        )
        self.lib.booth.create_in_cluster.assert_called_once_with(
            "1.2.3.4",
            instance_name=None,
            allow_absent_resource_agent=False,
        )

    def test_lib_call_full(self):
        booth_cmd.create_in_cluster(
            self.lib,
            ["ip", "1.2.3.4"],
            dict_to_modifiers(dict(name="my_booth", force=True)),
        )
        self.lib.booth.create_in_cluster.assert_called_once_with(
            "1.2.3.4",
            instance_name="my_booth",
            allow_absent_resource_agent=True,
        )


class RemoveFromClusterBase:
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth", "cluster", "env", "resource"])
        self.lib.env = mock.Mock(spec_set=["report_processor"])
        self.lib.booth = mock.Mock(
            spec_set=["get_resource_ids_from_cluster", "remove_from_cluster"]
        )
        self.lib.cluster = mock.Mock(spec_set=["wait_for_pcmk_idle"])
        self.lib.resource = mock.Mock(spec_set=["stop"])

    def _call_cmd(self, argv, modifiers=None):
        booth_cmd.remove_from_cluster(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def assert_lib_calls(self, expected_calls):
        self.lib.assert_has_calls(expected_calls)
        self.assertEqual(len(self.lib.mock_calls), len(expected_calls))

    def test_args(self, mock_process_library_reports, mock_reports):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["A"])
        self.assertIsNone(cm.exception.message)
        self.assert_lib_calls([])
        mock_reports.assert_not_called()
        mock_process_library_reports.assert_not_called()

    def test_no_name(self, mock_process_library_reports, mock_reports):
        mock_reports.return_value = [INFO_REPORT]
        self._call_cmd([])

        self.assert_lib_calls([mock.call.booth.remove_from_cluster(None)])
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=False
        )

    def test_name(self, mock_process_library_reports, mock_reports):
        mock_reports.return_value = [INFO_REPORT]
        self._call_cmd([], {"name": "A"})

        self.assert_lib_calls([mock.call.booth.remove_from_cluster("A")])
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=False
        )

    def test_with_debug(self, mock_process_library_reports, mock_reports):
        mock_reports.return_value = [INFO_REPORT]
        self._call_cmd([], {"debug": True})

        self.assert_lib_calls([mock.call.booth.remove_from_cluster(None)])
        mock_process_library_reports.assert_called_once_with(
            [INFO_REPORT], include_debug=True
        )

    def test_dont_stop_me_now(self, mock_process_library_reports, mock_reports):
        self._call_cmd([], {"no-stop": True})

        self.assert_lib_calls([mock.call.booth.remove_from_cluster(None, [])])
        mock_reports.assert_not_called()
        mock_process_library_reports.assert_not_called()

    def test_force_dont_stop_me_now(
        self, mock_process_library_reports, mock_reports
    ):
        self._call_cmd([], {"force": True, "no-stop": True})

        self.assert_lib_calls(
            [mock.call.booth.remove_from_cluster(None, [reports.codes.FORCE])]
        )
        mock_reports.assert_not_called()
        mock_process_library_reports.assert_not_called()

    def test_remove_not_stopped(
        self, mock_process_library_reports, mock_reports
    ):
        booth_name = None
        force_flags = []
        resource_ids = ["R1", "R2"]
        self.lib.booth.remove_from_cluster.side_effect = [LibraryError(), None]
        self.lib.booth.get_resource_ids_from_cluster.return_value = resource_ids
        mock_reports.return_value = [UNSTOPPED_RESOURCES_ERROR_REPORT]

        self._call_cmd([])

        self.assert_lib_calls(
            [
                mock.call.booth.remove_from_cluster(booth_name),
                mock.call.booth.get_resource_ids_from_cluster(booth_name),
                mock.call.resource.stop(resource_ids, force_flags),
                mock.call.cluster.wait_for_pcmk_idle(None),
                mock.call.booth.remove_from_cluster(booth_name, force_flags),
            ]
        )
        mock_process_library_reports.assert_not_called()

    def test_remove_more_errors(
        self, mock_process_library_reports, mock_reports
    ):
        self.lib.booth.remove_from_cluster.side_effect = [LibraryError(), None]
        mock_reports.return_value = [
            BOOTH_NOT_IN_CIB_ERROR,
            UNSTOPPED_RESOURCES_ERROR_REPORT,
        ]

        self.assertRaises(LibraryError, lambda: self._call_cmd([]))

        self.assert_lib_calls([mock.call.booth.remove_from_cluster(None)])
        mock_process_library_reports.assert_called_once_with(
            [BOOTH_NOT_IN_CIB_ERROR], include_debug=False, exit_on_error=False
        )

    def test_remove_more_errors_debug(
        self, mock_process_library_reports, mock_reports
    ):
        self.lib.booth.remove_from_cluster.side_effect = [LibraryError(), None]
        mock_reports.return_value = [
            BOOTH_NOT_IN_CIB_ERROR,
            UNSTOPPED_RESOURCES_ERROR_REPORT,
        ]

        self.assertRaises(
            LibraryError, lambda: self._call_cmd([], {"debug": True})
        )
        self.assert_lib_calls([mock.call.booth.remove_from_cluster(None)])
        mock_process_library_reports.assert_called_once_with(
            [BOOTH_NOT_IN_CIB_ERROR], include_debug=True, exit_on_error=False
        )

    def test_mutually_exclusive_options(
        self, mock_process_library_reports, mock_reports
    ):
        with self.assertRaises(CmdLineInputError) as cm:
            booth_cmd.remove_from_cluster(
                self.lib,
                [],
                InputModifiers({"-f": "foo", "--no-stop": True}),
            )
        self.assertEqual(
            cm.exception.message, "Only one of '--no-stop', '-f' can be used"
        )
        self.assert_lib_calls([])
        mock_reports.assert_not_called()
        mock_process_library_reports.assert_not_called()


@mock.patch(
    "pcs.common.reports.processor.ReportProcessorInMemory.reports",
    new_callable=mock.PropertyMock,
)
@mock.patch("pcs.cli.booth.command.process_library_reports")
class RemoveFromCluster(RemoveFromClusterBase, TestCase):
    @mock.patch("pcs.cli.booth.command.deprecation_warning")
    def test_remove_force(
        self, mock_deprecation_warning, mock_process_lib_reports, mock_reports
    ):
        self._call_cmd([], {"force": True})

        self.assert_lib_calls(
            [mock.call.booth.remove_from_cluster(None, [reports.codes.FORCE])]
        )
        mock_deprecation_warning.assert_called_once()
        mock_reports.assert_not_called()
        mock_process_lib_reports.assert_not_called()


@mock.patch(
    "pcs.common.reports.processor.ReportProcessorInMemory.reports",
    new_callable=mock.PropertyMock,
)
@mock.patch("pcs.cli.booth.command.process_library_reports")
class RemoveFromClusterFuture(RemoveFromClusterBase, TestCase):
    def _call_cmd(self, argv, modifiers=None):
        default_modifiers = {"future": True}
        booth_cmd.remove_from_cluster(
            self.lib,
            argv,
            dict_to_modifiers(
                modifiers | default_modifiers
                if modifiers
                else default_modifiers
            ),
        )

    def test_remove_force_more_errors_forceable(
        self, mock_process_lib_reports, mock_reports
    ):
        booth_name = None
        force_flags = [reports.codes.FORCE]
        resource_ids = ["R1"]
        self.lib.booth.remove_from_cluster.side_effect = [LibraryError(), None]
        self.lib.booth.get_resource_ids_from_cluster.return_value = resource_ids
        mock_reports.return_value = [
            FORCEABLE_ERROR_REPORT,
            UNSTOPPED_RESOURCES_ERROR_REPORT,
        ]

        self._call_cmd([], {"force": True})
        self.assert_lib_calls(
            [
                mock.call.booth.remove_from_cluster(booth_name),
                mock.call.booth.get_resource_ids_from_cluster(booth_name),
                mock.call.resource.stop(resource_ids, force_flags),
                mock.call.cluster.wait_for_pcmk_idle(None),
                mock.call.booth.remove_from_cluster(booth_name, force_flags),
            ]
        )
        mock_process_lib_reports.assert_not_called()

    def test_remove_force_more_errors_not_forceable(
        self, mock_process_lib_reports, mock_reports
    ):
        self.lib.booth.remove_from_cluster.side_effect = [LibraryError(), None]
        self.lib.booth.get_resource_ids_from_cluster.return_value = ["R1"]
        mock_reports.return_value = [
            FORCEABLE_ERROR_REPORT,
            BOOTH_NOT_IN_CIB_ERROR,
            UNSTOPPED_RESOURCES_ERROR_REPORT,
        ]

        self.assertRaises(
            LibraryError,
            lambda: self._call_cmd([], {"force": True}),
        )
        self.assert_lib_calls([mock.call.booth.remove_from_cluster(None)])
        mock_process_lib_reports.assert_called_once_with(
            [
                reports.item.ReportItem.warning(
                    reports.messages.BoothNotExistsInCib(name="booth")
                ),
                BOOTH_NOT_IN_CIB_ERROR,
            ],
            include_debug=False,
            exit_on_error=False,
        )


class TicketGrantTest(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["ticket_grant"])

    def test_lib_call_minimal(self):
        booth_cmd.ticket_grant(self.lib, ["ticketA"], dict_to_modifiers({}))
        self.lib.booth.ticket_grant.assert_called_once_with(
            "ticketA",
            instance_name=None,
            site_ip=None,
        )

    def test_lib_call_full(self):
        booth_cmd.ticket_grant(
            self.lib,
            ["ticketA", "1.2.3.4"],
            dict_to_modifiers(dict(name="my_booth")),
        )
        self.lib.booth.ticket_grant.assert_called_once_with(
            "ticketA",
            instance_name="my_booth",
            site_ip="1.2.3.4",
        )


class TicketRevokeTest(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["ticket_revoke"])

    def test_lib_call_minimal(self):
        booth_cmd.ticket_revoke(self.lib, ["ticketA"], dict_to_modifiers({}))
        self.lib.booth.ticket_revoke.assert_called_once_with(
            "ticketA",
            instance_name=None,
            site_ip=None,
        )

    def test_lib_call_full(self):
        booth_cmd.ticket_revoke(
            self.lib,
            ["ticketA", "1.2.3.4"],
            dict_to_modifiers(dict(name="my_booth")),
        )
        self.lib.booth.ticket_revoke.assert_called_once_with(
            "ticketA",
            instance_name="my_booth",
            site_ip="1.2.3.4",
        )


# disable printig the booth config so it won't break tests output
@mock.patch("pcs.cli.booth.command.print", new=lambda x: x)
class ConfigTest(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["config_text"])

    def test_too_many_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            booth_cmd.config_show(
                self.lib, ["aaa", "bbb"], dict_to_modifiers({})
            )
        self.assertIsNone(cm.exception.message)
        self.lib.booth.config_text.assert_not_called()

    def test_lib_call_minimal(self):
        booth_cmd.config_show(self.lib, [], dict_to_modifiers({}))
        self.lib.booth.config_text.assert_called_once_with(
            instance_name=None,
            node_name=None,
        )

    def test_lib_call_full(self):
        booth_cmd.config_show(
            self.lib,
            ["node1"],
            dict_to_modifiers(
                {
                    "name": "my_booth",
                    "request-timeout": "10",
                }
            ),
        )
        self.lib.booth.config_text.assert_called_once_with(
            instance_name="my_booth",
            node_name="node1",
        )


class Restart(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["restart"])

    def test_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            booth_cmd.restart(self.lib, ["something"], dict_to_modifiers({}))
        self.assertIsNone(cm.exception.message)
        self.lib.booth.restart.assert_not_called()

    def test_lib_call_minimal(self):
        booth_cmd.restart(self.lib, [], dict_to_modifiers({}))
        self.lib.booth.restart.assert_called_once_with(
            instance_name=None, allow_multiple=False
        )

    def test_lib_call_full(self):
        booth_cmd.restart(
            self.lib, [], dict_to_modifiers(dict(name="my_booth", force=True))
        )
        self.lib.booth.restart.assert_called_once_with(
            instance_name="my_booth", allow_multiple=True
        )


class Sync(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["config_sync"])

    def test_lib_call_minimal(self):
        booth_cmd.sync(self.lib, [], dict_to_modifiers({}))
        self.lib.booth.config_sync.assert_called_once_with(
            instance_name=None,
            skip_offline_nodes=False,
        )

    def test_lib_call_full(self):
        booth_cmd.sync(
            self.lib,
            [],
            dict_to_modifiers(
                {
                    "name": "my_booth",
                    "request-timeout": "10",
                    "skip-offline": True,
                    "booth-conf": "C",
                    "booth-key": "K",
                }
            ),
        )
        self.lib.booth.config_sync.assert_called_once_with(
            instance_name="my_booth",
            skip_offline_nodes=True,
        )


class BoothServiceTestMixin:
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])

    def test_lib_call_minimal(self):
        self.cli_cmd(self.lib, [], dict_to_modifiers({}))
        self.lib_cmd.assert_called_once_with(
            instance_name=None,
        )

    def test_lib_call_full(self):
        self.cli_cmd(self.lib, [], dict_to_modifiers(dict(name="my_booth")))
        self.lib_cmd.assert_called_once_with(
            instance_name="my_booth",
        )


class Enable(BoothServiceTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.lib.booth = mock.Mock(spec_set=["enable_booth"])
        self.cmd_label = "enable"
        self.lib_cmd = self.lib.booth.enable_booth
        self.cli_cmd = booth_cmd.enable


class Disable(BoothServiceTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.lib.booth = mock.Mock(spec_set=["disable_booth"])
        self.cmd_label = "disable"
        self.lib_cmd = self.lib.booth.disable_booth
        self.cli_cmd = booth_cmd.disable


class Start(BoothServiceTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.lib.booth = mock.Mock(spec_set=["start_booth"])
        self.cmd_label = "start"
        self.lib_cmd = self.lib.booth.start_booth
        self.cli_cmd = booth_cmd.start


class Stop(BoothServiceTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.lib.booth = mock.Mock(spec_set=["stop_booth"])
        self.cmd_label = "stop"
        self.lib_cmd = self.lib.booth.stop_booth
        self.cli_cmd = booth_cmd.stop


class Pull(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["pull_config"])

    def test_lib_call_minimal(self):
        booth_cmd.pull(self.lib, ["node1"], dict_to_modifiers({}))
        self.lib.booth.pull_config.assert_called_once_with(
            "node1",
            instance_name=None,
        )

    def test_lib_call_full(self):
        booth_cmd.pull(
            self.lib,
            ["node1"],
            dict_to_modifiers(
                {
                    "name": "my_booth",
                    "request-timeout": "10",
                }
            ),
        )
        self.lib.booth.pull_config.assert_called_once_with(
            "node1",
            instance_name="my_booth",
        )


# disable printig the booth status so it won't break tests output
@mock.patch("pcs.cli.booth.command.print", new=lambda x: x)
class Status(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["get_status"])
        self.lib.booth.get_status.return_value = {
            "ticket": "ticket_status",
            "peers": "peers_status",
            "daemon": "daemon_status",
        }

    def test_lib_call_minimal(self):
        booth_cmd.status(self.lib, [], dict_to_modifiers({}))
        self.lib.booth.get_status.assert_called_once_with(
            instance_name=None,
        )

    def test_lib_call_full(self):
        booth_cmd.status(self.lib, [], dict_to_modifiers(dict(name="my_booth")))
        self.lib.booth.get_status.assert_called_once_with(
            instance_name="my_booth",
        )


class TicketCleanup(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(
            spec_set=["ticket_cleanup", "ticket_cleanup_auto"]
        )

    def test_auto_minimal(self):
        booth_cmd.ticket_cleanup(self.lib, [], dict_to_modifiers({}))
        self.lib.booth.ticket_cleanup_auto.assert_called_once_with(
            instance_name=None
        )
        self.lib.booth.ticket_cleanup.assert_not_called()

    def test_auto_full(self):
        booth_cmd.ticket_cleanup(
            self.lib, [], dict_to_modifiers(dict(name="my_booth"))
        )
        self.lib.booth.ticket_cleanup_auto.assert_called_once_with(
            instance_name="my_booth"
        )
        self.lib.booth.ticket_cleanup.assert_not_called()

    def test_with_ticket_minimal(self):
        booth_cmd.ticket_cleanup(
            self.lib,
            ["ticketA"],
            dict_to_modifiers({}),
        )
        self.lib.booth.ticket_cleanup.assert_called_once_with("ticketA")
        self.lib.booth.ticket_cleanup_auto.assert_not_called()


class TicketUnstandby(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["ticket_unstandby"])

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            booth_cmd.ticket_unstandby(self.lib, [], dict_to_modifiers({}))
        self.assertIsNone(cm.exception.message)
        self.lib.booth.ticket_unstandby.assert_not_called()

    def test_too_many_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            booth_cmd.ticket_unstandby(
                self.lib, ["a", "b"], dict_to_modifiers({})
            )
        self.assertIsNone(cm.exception.message)
        self.lib.booth.ticket_unstandby.assert_not_called()

    def test_call_minimal(self):
        booth_cmd.ticket_unstandby(
            self.lib, ["my_ticket"], dict_to_modifiers({})
        )
        self.lib.booth.ticket_unstandby.assert_called_once_with("my_ticket")


class TicketStandby(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["booth"])
        self.lib.booth = mock.Mock(spec_set=["ticket_standby"])

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            booth_cmd.ticket_standby(self.lib, [], dict_to_modifiers({}))
        self.assertIsNone(cm.exception.message)
        self.lib.booth.ticket_standby.assert_not_called()

    def test_too_many_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            booth_cmd.ticket_standby(
                self.lib, ["a", "b"], dict_to_modifiers({})
            )
        self.assertIsNone(cm.exception.message)
        self.lib.booth.ticket_standby.assert_not_called()

    def test_call_minimal(self):
        booth_cmd.ticket_standby(self.lib, ["my_ticket"], dict_to_modifiers({}))
        self.lib.booth.ticket_standby.assert_called_once_with("my_ticket")



import os
from typing import Iterable, Optional

from pcs import settings
from pcs.common.node_communicator import (
    CommunicatorLoggerInterface,
    Destination,
    Request,
    Response,
)
from pcs.common.reports import ReportItem, ReportProcessor, messages
from pcs.lib.external import is_proxy_set


def _get_port(port: Optional[int]) -> int:
    return port if port is not None else settings.pcsd_default_port


class CommunicatorLogger(CommunicatorLoggerInterface):
    def __init__(self, reporters: Iterable[ReportProcessor]):
        self._reporters = reporters

    def _log_report_to_all_reporters(self, msg: ReportItem) -> None:
        for reporter in self._reporters:
            reporter.report(msg)

    def log_request_start(self, request: Request) -> None:
        self._log_report_to_all_reporters(
            ReportItem.debug(
                messages.NodeCommunicationStarted(request.url, request.data)
            )
        )

    def log_response(self, response: Response) -> None:
        if response.was_connected:
            self._log_response_successful(response)
        else:
            self._log_response_failure(response)
        self._log_debug(response)

    def _log_response_successful(self, response: Response) -> None:
        self._log_report_to_all_reporters(
            ReportItem.debug(
                messages.NodeCommunicationFinished(
                    response.request.url,
                    response.response_code,  # type: ignore
                    response.data,
                )
            )
        )

    def _log_response_failure(self, response: Response) -> None:
        self._log_report_to_all_reporters(
            ReportItem.debug(
                messages.NodeCommunicationNotConnected(
                    response.request.host_label, response.error_msg or ""
                )
            )
        )
        if is_proxy_set(os.environ):
            self._log_report_to_all_reporters(
                ReportItem.warning(
                    messages.NodeCommunicationProxyIsSet(
                        response.request.host_label, response.request.dest.addr
                    )
                )
            )

    def _log_debug(self, response: Response) -> None:
        self._log_report_to_all_reporters(
            ReportItem.debug(
                messages.NodeCommunicationDebugInfo(
                    response.request.url, response.debug
                )
            )
        )

    def log_retry(self, response: Response, previous_dest: Destination) -> None:
        self._log_report_to_all_reporters(
            ReportItem.warning(
                messages.NodeCommunicationRetrying(
                    response.request.host_label,
                    previous_dest.addr,
                    str(_get_port(previous_dest.port)),
                    response.request.dest.addr,
                    str(_get_port(response.request.dest.port)),
                    response.request.url,
                )
            )
        )

    def log_no_more_addresses(self, response: Response) -> None:
        self._log_report_to_all_reporters(
            ReportItem.warning(
                messages.NodeCommunicationNoMoreAddresses(
                    response.request.host_label, response.request.url
                )
            )
        )



from unittest import TestCase

from pcs.cli.constraint.output import filter_constraints_by_rule_expired_status

from pcs_test.tools.constraints_dto import get_all_constraints
from pcs_test.tools.custom_mock import RuleInEffectEvalMock


class FilterConstraintsByRuleExpiredStatus(TestCase):
    def setUp(self):
        self.constraint_dtos_with_expired = get_all_constraints(
            RuleInEffectEvalMock({}), include_expired=True
        )
        self.constraint_dtos_without_expired = get_all_constraints(
            RuleInEffectEvalMock({}), include_expired=False
        )

    def test_include_expired(self):
        self.assertEqual(
            filter_constraints_by_rule_expired_status(
                self.constraint_dtos_with_expired, include_expired=True
            ),
            self.constraint_dtos_with_expired,
        )

    def test_do_not_include_expired(self):
        self.assertEqual(
            filter_constraints_by_rule_expired_status(
                self.constraint_dtos_with_expired, include_expired=False
            ),
            self.constraint_dtos_without_expired,
        )



from unittest import TestCase, mock

from pcs.cli.cluster.report_processor import NodeRemoveRemoteReportProcessor
from pcs.common import reports


@mock.patch("pcs.cli.cluster.report_processor.print_report")
class NodeRemoveRemote(TestCase):
    REPORT = reports.ReportItem.info(reports.messages.NoActionNecessary())
    DEBUG_REPORT = reports.ReportItem.debug(
        reports.messages.NoActionNecessary()
    )
    FIRST_SERVICE_COMMAND_REPORT = reports.ReportItem.info(
        reports.messages.ServiceCommandsOnNodesStarted()
    )

    def setUp(self):
        self.report_processor = NodeRemoveRemoteReportProcessor()

    def test_saves_reports_in_memory(self, mock_print_report):
        self.report_processor.report(self.REPORT)
        self.assertEqual(self.report_processor.reports, [self.REPORT])
        self.assertFalse(self.report_processor.already_reported_to_console)
        mock_print_report.assert_not_called()

        self.report_processor.report(self.REPORT)
        self.assertEqual(
            self.report_processor.reports, [self.REPORT, self.REPORT]
        )
        self.assertFalse(self.report_processor.already_reported_to_console)
        mock_print_report.assert_not_called()

    def test_switches_to_printing(self, mock_print_report):
        self.report_processor.report_list([self.REPORT, self.REPORT])
        self.assertEqual(
            self.report_processor.reports, [self.REPORT, self.REPORT]
        )
        self.assertFalse(self.report_processor.already_reported_to_console)
        mock_print_report.assert_not_called()

        self.report_processor.report(self.FIRST_SERVICE_COMMAND_REPORT)
        self.assertEqual(self.report_processor.reports, [])
        self.assertTrue(self.report_processor.already_reported_to_console)
        self.assertEqual(mock_print_report.call_count, 3)
        mock_print_report.assert_has_calls(
            [
                mock.call(self.REPORT.to_dto()),
                mock.call(self.REPORT.to_dto()),
                mock.call(self.FIRST_SERVICE_COMMAND_REPORT.to_dto()),
            ]
        )

        self.report_processor.report(self.REPORT)
        self.assertEqual(self.report_processor.reports, [])
        self.assertEqual(mock_print_report.call_count, 4)
        mock_print_report.assert_has_calls(
            [
                mock.call(self.REPORT.to_dto()),
                mock.call(self.REPORT.to_dto()),
                mock.call(self.FIRST_SERVICE_COMMAND_REPORT.to_dto()),
                mock.call(self.REPORT.to_dto()),
            ]
        )

    def test_filter_debug(self, mock_print_report):
        self.report_processor.report_list([self.REPORT, self.DEBUG_REPORT])
        self.assertEqual(self.report_processor.reports, [self.REPORT])
        mock_print_report.assert_not_called()

    def test_keep_debug(self, mock_print_report):
        self.report_processor = NodeRemoveRemoteReportProcessor(True)

        self.report_processor.report_list([self.REPORT, self.DEBUG_REPORT])
        self.assertEqual(
            self.report_processor.reports, [self.REPORT, self.DEBUG_REPORT]
        )
        mock_print_report.assert_not_called()



from pcs.common.reports import codes as report_codes

from pcs_test.tools import fixture

OFFLINE_ERROR_MSG = "Could not resolve host"


class EnvConfigMixin:
    def __init__(self, call_collection, wrap_helper, config):
        del wrap_helper, call_collection
        self.config = config

    def destroy_pacemaker_remote(
        self, label=None, dest_list=None, result=None, **kwargs
    ):
        if kwargs.get("was_connected", True):
            result = (
                result
                if result is not None
                else {
                    "code": "success",
                    "message": "",
                }
            )

            kwargs["results"] = {
                "pacemaker_remote stop": result,
                "pacemaker_remote disable": result,
            }
        elif result is not None:
            raise AssertionError(
                "Keyword 'result' makes no sense with 'was_connected=False'"
            )

        if label or dest_list:
            if kwargs.get("communication_list"):
                raise AssertionError(
                    "Keywords 'label' and 'dest_list' makes no sense with"
                    " 'communication_list != None'"
                )
            kwargs["communication_list"] = [
                dict(label=label, dest_list=dest_list)
            ]

        self.config.http.manage_services(
            action_map={
                "pacemaker_remote stop": {
                    "type": "service_command",
                    "service": "pacemaker_remote",
                    "command": "stop",
                },
                "pacemaker_remote disable": {
                    "type": "service_command",
                    "service": "pacemaker_remote",
                    "command": "disable",
                },
            },
            **kwargs,
        )

    def remove_authkey(self, communication_list, result=None, **kwargs):
        if kwargs.get("was_connected", True):
            result = (
                result
                if result is not None
                else {
                    "code": "deleted",
                    "message": "",
                }
            )

            kwargs["results"] = {"pacemaker_remote authkey": result}
        elif result is not None:
            raise AssertionError(
                "Keyword 'result' makes no sense with 'was_connected=False'"
            )
        self.config.http.remove_file(
            communication_list=communication_list,
            files={
                "pacemaker_remote authkey": {
                    "type": "pcmk_remote_authkey",
                }
            },
            **kwargs,
        )


def base_reports_for_host(host):
    return (
        fixture.ReportSequenceBuilder()
        .info(
            report_codes.SERVICE_COMMANDS_ON_NODES_STARTED,
            action_list=[
                "pacemaker_remote stop",
                "pacemaker_remote disable",
            ],
            node_list=[host],
            _name="pcmk_remote_disable_stop_started",
        )
        .info(
            report_codes.SERVICE_COMMAND_ON_NODE_SUCCESS,
            service_command_description="pacemaker_remote disable",
            node=host,
            _name="pcmk_remote_disable_success",
        )
        .info(
            report_codes.SERVICE_COMMAND_ON_NODE_SUCCESS,
            service_command_description="pacemaker_remote stop",
            node=host,
            _name="pcmk_remote_stop_success",
        )
        .info(
            report_codes.FILES_REMOVE_FROM_NODES_STARTED,
            file_list=["pacemaker authkey"],
            node_list=[host],
            _name="authkey_remove_started",
        )
        .info(
            report_codes.FILE_REMOVE_FROM_NODE_SUCCESS,
            file_description="pacemaker authkey",
            node=host,
            _name="authkey_remove_success",
        )
        .fixtures
    )


def report_remove_file_connection_failed(node):
    return fixture.error(
        report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
        force_code=report_codes.SKIP_OFFLINE_NODES,
        command="remote/remove_file",
        reason=OFFLINE_ERROR_MSG,
        node=node,
    )


def report_authkey_remove_failed(node):
    return fixture.error(
        report_codes.FILE_REMOVE_FROM_NODE_ERROR,
        force_code=report_codes.FORCE,
        reason="Access denied",
        file_description="pacemaker authkey",
        node=node,
    )


def report_pcmk_remote_disable_failed(node):
    return fixture.error(
        report_codes.SERVICE_COMMAND_ON_NODE_ERROR,
        force_code=report_codes.FORCE,
        reason="Operation failed.",
        service_command_description="pacemaker_remote disable",
        node=node,
    )


def report_pcmk_remote_stop_failed(node):
    return fixture.error(
        report_codes.SERVICE_COMMAND_ON_NODE_ERROR,
        force_code=report_codes.FORCE,
        reason="Operation failed.",
        service_command_description="pacemaker_remote stop",
        node=node,
    )



from typing import (
    Final,
    NewType,
    Union,
)

FencingTargetType = NewType("FencingTargetType", str)
FencingTargetValue = Union[str, tuple[str, str]]

TARGET_TYPE_NODE: Final = FencingTargetType("node")
TARGET_TYPE_REGEXP: Final = FencingTargetType("regexp")
TARGET_TYPE_ATTRIBUTE: Final = FencingTargetType("attribute")



from typing import Any

import pcs.cli.constraint_colocation.command as colocation_command
from pcs import (
    constraint,
    usage,
)
from pcs.cli.common.errors import (
    CmdLineInputError,
    raise_command_removed,
    raise_command_replaced,
)
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
)
from pcs.cli.common.routing import create_router
from pcs.cli.constraint import command as constraint_command
from pcs.cli.constraint.location import command as location_command
from pcs.cli.constraint_ticket import command as ticket_command
from pcs.utils import exit_on_cmdline_input_error


def constraint_location_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    sub_cmd = "config" if not argv else argv.pop(0)

    try:
        if sub_cmd == "add":
            constraint.location_add(lib, argv, modifiers)
        elif sub_cmd in ["remove", "delete"]:
            location_command.remove(lib, argv, modifiers)
        elif sub_cmd == "show":
            raise_command_replaced(
                ["pcs constraint location config"], pcs_version="0.12"
            )
        elif sub_cmd == "config":
            constraint.location_config_cmd(lib, argv, modifiers)
        elif len(argv) >= 2:
            if argv[0] == "rule":
                location_command.create_with_rule(
                    lib, [sub_cmd] + argv, modifiers
                )
            else:
                constraint.location_prefer(lib, [sub_cmd] + argv, modifiers)
        else:
            raise CmdLineInputError()
    except CmdLineInputError as e:
        exit_on_cmdline_input_error(e, "constraint", ["location", sub_cmd])


constraint_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.constraint(argv)),
        "location": constraint_location_cmd,
        "order": constraint.constraint_order_cmd,
        "ticket": create_router(
            {
                "set": ticket_command.create_with_set,
                "add": ticket_command.add,
                "delete": ticket_command.remove,
                "remove": ticket_command.remove,
                "show": lambda lib, argv, modifiers: raise_command_replaced(
                    ["pcs constraint ticket config"], pcs_version="0.12"
                ),
                "config": ticket_command.config_cmd,
            },
            ["constraint", "ticket"],
            default_cmd="config",
        ),
        "colocation": create_router(
            {
                "add": constraint.colocation_add,
                "remove": colocation_command.remove,
                "delete": colocation_command.remove,
                "set": colocation_command.create_with_set,
                "show": lambda lib, argv, modifiers: raise_command_replaced(
                    ["pcs constraint colocation config"], pcs_version="0.12"
                ),
                "config": colocation_command.config_cmd,
            },
            ["constraint", "colocation"],
            default_cmd="config",
        ),
        "remove": constraint_command.remove,
        "delete": constraint_command.remove,
        "show": lambda lib, argv, modifiers: raise_command_replaced(
            ["pcs constraint config"], pcs_version="0.12"
        ),
        "list": lambda lib, argv, modifiers: raise_command_replaced(
            ["pcs constraint config"], pcs_version="0.12"
        ),
        "config": constraint.config_cmd,
        "ref": constraint.ref,
        "rule": lambda lib, argv, modifiers: raise_command_removed(
            pcs_version="0.12"
        ),
    },
    ["constraint"],
    default_cmd="config",
)



from typing import Any, Optional, cast

from lxml import etree
from lxml.etree import _Element

from pcs.common import reports
from pcs.common.pacemaker.alert import (
    CibAlertDto,
    CibAlertRecipientDto,
    CibAlertSelectAttributeDto,
    CibAlertSelectDto,
)
from pcs.lib.cib import nvpair_multi, rule
from pcs.lib.cib.const import TAG_ALERT, TAG_RECIPIENT
from pcs.lib.cib.nvpair import get_nvset
from pcs.lib.cib.tools import (
    ElementSearcher,
    IdProvider,
    create_subelement_id,
    get_alerts,
)
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import validate_id_reports
from pcs.lib.xml_tools import (
    get_sub_element,
    remove_one_element,
    update_attribute_remove_empty,
)


def get_all_alert_elements(tree: _Element) -> list[_Element]:
    return tree.findall(TAG_ALERT)


def find_alert(context_el: _Element, alert_id: str) -> _Element:
    searcher = ElementSearcher(TAG_ALERT, alert_id, context_el)
    found_element = searcher.get_element()
    if found_element is not None:
        return found_element
    raise LibraryError(*searcher.get_errors())


def find_recipient(context_el: _Element, recipient_id: str) -> _Element:
    searcher = ElementSearcher(TAG_RECIPIENT, recipient_id, context_el)
    found_element = searcher.get_element()
    if found_element is not None:
        return found_element
    raise LibraryError(*searcher.get_errors())


def _update_optional_attribute(
    element: _Element, attribute: str, value: Optional[str]
) -> None:
    """
    Set value of an optional attribute, remove the attribute on empty value

    element -- element to be updated
    attribute -- attribute to be updated
    value -- new value of the attribute
    """
    if value is None:
        return
    update_attribute_remove_empty(element, attribute, value)


def _validate_recipient_value_is_unique(
    alert_el: _Element,
    recipient_value: str,
    recipient_id: Optional[str] = None,
    allow_duplicity: bool = False,
) -> reports.ReportItemList:
    """
    Validate that the recipient_value is unique in the specified alert

    alert_el -- alert
    recipient_value -- recipient value
    recipient_id -- id of the recipient to which the value belongs
    allow_duplicity -- if True, report a warning if the value already exists
    """
    report_list: reports.ReportItemList = []
    recipient_list = alert_el.xpath(
        "./recipient[@value=$value and @id!=$id]",
        value=recipient_value,
        id=recipient_id or "",
    )
    if recipient_list:
        report_list.append(
            reports.ReportItem(
                severity=reports.item.get_severity(
                    reports.codes.FORCE,
                    allow_duplicity,
                ),
                message=reports.messages.CibAlertRecipientAlreadyExists(
                    str(alert_el.attrib["id"]),
                    recipient_value,
                ),
            )
        )
    return report_list


def validate_create_alert(
    id_provider: IdProvider,
    # should be str, see lib.commands.alert.create_alert
    path: Optional[str],
    alert_id: Optional[str] = None,
) -> reports.ReportItemList:
    """
    validate new alert creation

    id_provider -- elements' ids generator
    path -- path to script
    alert_id -- id of new alert or None
    """
    report_list: reports.ReportItemList = []
    if not path:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(["path"])
            )
        )
    if alert_id:
        report_list.extend(
            validate_id_reports(alert_id, description="alert-id")
        )
        report_list.extend(id_provider.book_ids(alert_id))
    return report_list


def create_alert(
    tree: _Element,
    id_provider: IdProvider,
    path: str,
    alert_id: Optional[str] = None,
    description: Optional[str] = None,
) -> _Element:
    """
    Create new alert element and return it

    tree -- cib etree node
    id_provider -- elements' ids generator
    path -- path to script
    alert_id -- id of new alert, it will be generated if it is None
    description -- description
    """
    if not alert_id:
        alert_id = id_provider.allocate_id(TAG_ALERT)

    alert_el = etree.SubElement(
        get_alerts(tree), TAG_ALERT, id=alert_id, path=path
    )
    if description:
        alert_el.set("description", description)

    return alert_el


def update_alert(
    tree: _Element,
    alert_id: str,
    path: Optional[str],
    description: Optional[str] = None,
) -> _Element:
    """
    Update existing alert. Return updated alert element.
    Raises LibraryError if alert with specified id doesn't exist.

    tree -- cib etree node
    alert_id -- id of alert to be updated
    path -- new value of path, stay unchanged if None
    description -- new value of description, stay unchanged if None, remove
        if empty
    """
    alert = find_alert(get_alerts(tree), alert_id)
    if path:
        alert.set("path", path)
    _update_optional_attribute(alert, "description", description)
    return alert


def remove_alert(tree: _Element, alert_id: str) -> None:
    """
    Remove alert with specified id.
    Raises LibraryError if alert with specified id doesn't exist.

    tree -- cib etree node
    alert_id -- id of alert which should be removed
    """
    remove_one_element(find_alert(get_alerts(tree), alert_id))


def validate_add_recipient(
    id_provider: IdProvider,
    alert_el: _Element,
    # should be str, see lib.commands.alert.add_recipient
    recipient_value: Optional[str],
    recipient_id: Optional[str] = None,
    allow_same_value: bool = False,
) -> reports.ReportItemList:
    """
    Validate adding a recipient to the specified alert

    id_provider -- elements' ids generator
    alert_el -- alert which should be the parent of the new recipient
    recipient_value -- value of the new recipient
    recipient_id -- id of the new recipient or None
    allow_same_value -- if True, unique recipient value is not required
    """
    report_list: reports.ReportItemList = []

    if not recipient_value:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(["value"])
            )
        )
    else:
        report_list.extend(
            _validate_recipient_value_is_unique(
                alert_el,
                recipient_value,
                recipient_id,
                allow_duplicity=allow_same_value,
            )
        )

    if recipient_id:
        report_list.extend(
            validate_id_reports(recipient_id, description="recipient-id")
        )
        report_list.extend(id_provider.book_ids(recipient_id))

    return report_list


def add_recipient(
    id_provider: IdProvider,
    alert_el: _Element,
    recipient_value: str,
    recipient_id: Optional[str] = None,
    description: Optional[str] = None,
) -> _Element:
    """
    Add new recipient to the specified alert, return added recipient element

    id_provider -- elements' ids generator
    alert_el -- alert which should be the parent of the new recipient
    recipient_value -- value of the new recipient
    recipient_id -- id of the new recipient or None
    description -- description of the new recipient
    """
    if not recipient_id:
        recipient_id = create_subelement_id(
            alert_el, TAG_RECIPIENT, id_provider
        )
    recipient_el = etree.SubElement(
        alert_el, TAG_RECIPIENT, id=recipient_id, value=recipient_value
    )
    if description:
        recipient_el.attrib["description"] = description
    return recipient_el


def validate_update_recipient(
    recipient_el: _Element,
    recipient_value: Optional[str] = None,
    allow_same_value: bool = False,
) -> reports.ReportItemList:
    """
    validate updating specified recipient

    recipient_el -- the recipient to be updated
    recipient_value -- new recipient value, stay unchanged if None
    allow_same_value -- if True, unique recipient value is not required
    """
    report_list: reports.ReportItemList = []

    if recipient_value is not None:
        if not recipient_value:
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.CibAlertRecipientValueInvalid(
                        recipient_value
                    )
                )
            )
        else:
            report_list.extend(
                _validate_recipient_value_is_unique(
                    cast(_Element, recipient_el.getparent()),
                    recipient_value,
                    str(recipient_el.attrib["id"]),
                    allow_duplicity=allow_same_value,
                )
            )

    return report_list


def update_recipient(
    recipient_el: _Element,
    recipient_value: Optional[str] = None,
    description: Optional[str] = None,
) -> _Element:
    """
    Update specified recipient. Returns updated recipient element.

    recipient_el -- the recipient to be updated
    recipient_value -- new recipient value, stay unchanged if None
    description -- description, if empty it will be removed, stay unchanged
        if None
    """
    if recipient_value is not None:
        recipient_el.set("value", recipient_value)
    _update_optional_attribute(recipient_el, "description", description)
    return recipient_el


def remove_recipient(tree: _Element, recipient_id: str) -> None:
    """
    Remove specified recipient.
    Raises LibraryError if recipient doesn't exist.

    tree -- cib etree node
    recipient_id -- id of recipient to be removed
    """
    remove_one_element(find_recipient(get_alerts(tree), recipient_id))


def _recipient_el_to_dto(
    recipient_el: _Element,
    rule_eval: Optional[rule.RuleInEffectEval] = None,
) -> CibAlertRecipientDto:
    if rule_eval is None:
        rule_eval = rule.RuleInEffectEvalDummy()
    return CibAlertRecipientDto(
        id=str(recipient_el.attrib["id"]),
        value=str(recipient_el.attrib["value"]),
        description=recipient_el.get("description"),
        meta_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                recipient_el, nvpair_multi.NVSET_META
            )
        ],
        instance_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                recipient_el, nvpair_multi.NVSET_INSTANCE
            )
        ],
    )


def _select_el_to_dto(select_el: _Element) -> CibAlertSelectDto:
    return CibAlertSelectDto(
        nodes=(select_el.find("select_nodes") is not None),
        fencing=(select_el.find("select_fencing") is not None),
        resources=(select_el.find("select_resources") is not None),
        attributes=(select_el.find("select_attributes") is not None),
        attributes_select=[
            CibAlertSelectAttributeDto(
                str(attr_el.attrib["id"]), str(attr_el.attrib["name"])
            )
            for attr_el in select_el.iterfind("select_attributes/attribute")
        ],
    )


def alert_el_to_dto(
    alert_el: _Element,
    rule_eval: Optional[rule.RuleInEffectEval] = None,
) -> CibAlertDto:
    if rule_eval is None:
        rule_eval = rule.RuleInEffectEvalDummy()
    select_el = alert_el.find("select")
    return CibAlertDto(
        id=str(alert_el.attrib["id"]),
        path=str(alert_el.attrib["path"]),
        description=alert_el.get("description"),
        recipients=[
            _recipient_el_to_dto(recipient_el, rule_eval)
            for recipient_el in alert_el.iterfind(TAG_RECIPIENT)
        ],
        select=_select_el_to_dto(select_el) if select_el is not None else None,
        meta_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                alert_el, nvpair_multi.NVSET_META
            )
        ],
        instance_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                alert_el, nvpair_multi.NVSET_INSTANCE
            )
        ],
    )


# DEPRECATED, used only in get_all_alerts_dict
def get_all_recipients_dict(alert: _Element) -> list[dict[str, Any]]:
    """
    Returns list of all recipient of specified alert. Format:
    [
        {
            "id": <id of recipient>,
            "value": <value of recipient>,
            "description": <recipient description>,
            "instance_attributes": <list of nvpairs>,
            "meta_attributes": <list of nvpairs>
        }
    ]

    alert -- parent element of recipients to return
    """
    return [
        {
            "id": recipient.get("id"),
            "value": recipient.get("value"),
            "description": recipient.get("description", ""),
            "instance_attributes": get_nvset(
                get_sub_element(recipient, "instance_attributes")
            ),
            "meta_attributes": get_nvset(
                get_sub_element(recipient, "meta_attributes")
            ),
        }
        for recipient in alert.findall("./recipient")
    ]


# DEPRECATED, use alert_el_to_dto + get_all_alert_elements
def get_all_alerts_dict(tree: _Element) -> list[dict[str, Any]]:
    """
    Returns list of all alerts specified in tree. Format:
    [
        {
            "id": <id of alert>,
            "path": <path to script>,
            "description": <alert description>,
            "instance_attributes": <list of nvpairs>,
            "meta_attributes": <list of nvpairs>,
            "recipients_list": <list of alert's recipients>
        }
    ]

    tree -- cib etree node
    """
    return [
        {
            "id": alert.get("id"),
            "path": alert.get("path"),
            "description": alert.get("description", ""),
            "instance_attributes": get_nvset(
                get_sub_element(alert, "instance_attributes")
            ),
            "meta_attributes": get_nvset(
                get_sub_element(alert, "meta_attributes")
            ),
            "recipient_list": get_all_recipients_dict(alert),
        }
        for alert in get_alerts(tree).findall("./alert")
    ]



from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
    group_by_keywords,
)
from pcs.cli.tag.output import print_config


def tag_create(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if len(argv) < 2:
        raise CmdLineInputError()
    tag_id, idref_list = argv[0], argv[1:]
    lib.tag.create(tag_id, idref_list)


def tag_config(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --output-format - supported formats: text, cmd, json
    """
    modifiers.ensure_only_supported("-f", "--output-format")
    tag_dto = lib.tag.get_config_dto(argv)
    print_config(tag_dto, modifiers)


def tag_remove(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if len(argv) < 1:
        raise CmdLineInputError()
    lib.tag.remove(argv)


def tag_update(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --after - place a reference id in a tag after the specified reference
        id in the tag
      * --before - place a reference id in a tag before the specified reference
        id in the tag
    """
    modifiers.ensure_only_supported("-f", "--after", "--before")
    if not argv:
        raise CmdLineInputError()
    tag_id = argv[0]
    parsed_args = group_by_keywords(argv[1:], ["add", "remove"])
    parsed_args.ensure_unique_keywords()
    no_add_remove_arguments = not (
        parsed_args.has_keyword("add") or parsed_args.has_keyword("remove")
    )
    no_add_id = parsed_args.has_empty_keyword("add")
    no_remove_id = parsed_args.has_empty_keyword("remove")
    if no_add_remove_arguments or no_add_id or no_remove_id:
        raise CmdLineInputError(
            show_both_usage_and_message=True,
            hint=("Specify at least one id for 'add' or 'remove' arguments."),
        )
    adjacent_idref = None
    after_adjacent = True
    if modifiers.is_specified("--after") and modifiers.is_specified("--before"):
        raise CmdLineInputError("Cannot specify both --before and --after")
    if modifiers.is_specified("--after"):
        adjacent_idref = modifiers.get("--after")
        after_adjacent = True
    elif modifiers.is_specified("--before"):
        adjacent_idref = modifiers.get("--before")
        after_adjacent = False
    lib.tag.update(
        tag_id,
        parsed_args.get_args_flat("add"),
        parsed_args.get_args_flat("remove"),
        adjacent_idref=adjacent_idref,
        put_after_adjacent=after_adjacent,
    )
